# -- Project information -----------------------------------------------------

project = "Grader Service"
copyright = "2024"
author = "TU Wien dataLAB"
# language = "fr"  # For testing language translations

master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "ablog",
    "myst_nb",
    "autodoc_traits",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "sphinx_examples",
    "sphinx_tabs.tabs",
    "sphinx_thebe",
    "sphinx_togglebutton",
    "sphinx.ext.todo",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

nitpick_ignore = [
    ("py:class", "docutils.nodes.document"),
    ("py:class", "docutils.parsers.rst.directives.body.Sidebar"),
]

suppress_warnings = ["myst.domains", "ref.ref"]

numfig = True

myst_enable_extensions = ["dollarmath", "amsmath", "deflist", "colon_fence"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_logo = "_static/assets/images/logo_name.png"
html_title = "Sphinx Book Theme"
html_copy_source = True
html_favicon = "_static/assets/images/logo_name.png"
html_last_updated_fmt = ""

html_sidebars = {
    "reference/blog/*": [
        "navbar-logo.html",
        "search-field.html",
        "ablog/postcard.html",
        "ablog/recentposts.html",
        "ablog/tagcloud.html",
        "ablog/categories.html",
        "ablog/archives.html",
        "sbt-sidebar-nav.html",
    ]
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
nb_execution_mode = "cache"

html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/TU-Wien-dataLAB/Grader-Service",
    "repository_branch": "master",
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "colab_url": "https://colab.research.google.com/",
        "deepnote_url": "https://deepnote.com/",
        "notebook_interface": "jupyterlab",
        "thebe": True,
    },
    "use_edit_page_button": True,
    "use_source_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "use_sidenotes": True,
    "show_toc_level": 3,
    "logo": {"image_dark": "_static/assets/images/logo_name.png"},
    "icon_links": [
        {
            "name": "GitHub Grader Service",
            "url": "https://github.com/TU-Wien-dataLAB/Grader-Service",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI Grader Service",
            "url": "https://pypi.org/project/grader-service/",
            "icon": "https://img.shields.io/pypi/v/grader-service",
            "type": "url",
        },
        {
            "name": "TU Wien dataLAB",
            "url": "https://www.it.tuwien.ac.at/en/services/network-and-servers/datalab",
            "icon": "_static/assets/images/tu-logo.svg",
            "type": "url",
        },
    ],
}

ogp_social_cards = {"image": "_static/assets/images/logo_name.png"}



import os
from typing import Optional

from jupyterhub.auth import Authenticator
from traitlets import log as traitlets_log

from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.auth.token import JupyterHubTokenAuthenticator

from grader_service.handlers.base_handler import BaseHandler
from grader_service.orm import User, Lecture
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope, Role


logger = traitlets_log.get_logger()

c.GraderService.service_host = "0.0.0.0"
c.GraderService.service_port = 4010
c.GraderService.grader_service_dir = "/app/service_dir"

# choose the right db_url based on the environment variable
database_type = os.getenv("DATABASE_TYPE")
db_url = ""
if database_type == "sqlite":
    db_url = "sqlite:///grader.db"
elif database_type == "postgres":
    db_url = "postgresql://postgres:postgres@postgres:5432/grader"
else:
    logger.error("Unknown database type")

c.GraderService.db_url = db_url

c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor
c.LocalAutogradeExecutor.default_cell_timeout = 200

# get rabbitmq username and password
rabbit_mq_username = os.getenv("RABBITMQ_GRADER_SERVICE_USERNAME")
rabbit_mq_password = os.getenv("RABBITMQ_GRADER_SERVICE_PASSWORD")

c.CeleryApp.conf = dict(
    broker_url=f"amqp://{rabbit_mq_username}:{rabbit_mq_password}@rabbitmq:5672/grader",
    result_backend="rpc://",
    task_serializer="json",
    result_serializer="json",
    accept_content=["json"],
    broker_connection_retry_on_startup=True,
)


c.GraderService.authenticator_class = JupyterHubTokenAuthenticator


def post_auth_hook(authenticator: Authenticator, handler: BaseHandler, authentication: dict):
    log = handler.log
    log.info("post_auth_hook started")

    session = handler.session
    groups: list[str] = authentication["groups"]

    username = authentication["name"]
    user_model: Optional[User] = session.query(User).filter(User.name == username).one_or_none()
    if user_model is None:
        user_model = User()
        user_model.name = username
        user_model.display_name = username
        session.add(user_model)
        session.commit()

    for group in groups:
        if ":" in group:
            split_group = group.split(":")
            lecture_code = split_group[0]
            scope = split_group[1]
            scope = Scope[scope]

            lecture = session.query(Lecture).filter(Lecture.code == lecture_code).one_or_none()
            if lecture is None:
                lecture = Lecture()
                lecture.code = lecture_code
                lecture.name = lecture_code
                lecture.state = LectureState.active
                lecture.deleted = DeleteState.active
                session.add(lecture)
                session.commit()

            role = (
                session.query(Role)
                .filter(Role.user_id == user_model.id, Role.lectid == lecture.id)
                .one_or_none()
            )
            if role is None:
                log.info(
                    "No role for user %s in lecture %s... creating role", username, lecture_code
                )
                role = Role(user_id=user_model.id, lectid=lecture.id, role=scope)
                session.add(role)
                session.commit()
            else:
                log.info(
                    "Found role %s for user %s in lecture %s... updating role to %s",
                    role.role.name,
                    username,
                    lecture_code,
                    scope.name,
                )
                role.role = scope
                session.commit()
        else:
            log.info("Found group that doesn't match schema. Ignoring " + group)

    return authentication


c.Authenticator.post_auth_hook = post_auth_hook

c.JupyterHubTokenAuthenticator.user_info_url = "http://hub:8080/hub/api/user"
c.Authenticator.admin_users = {"admin"}
c.Authenticator.allow_all = True



from typing import Union

from celery import Celery
from sqlalchemy.orm import scoped_session
from traitlets import Dict
from traitlets.config import SingletonConfigurable


class CeleryApp(SingletonConfigurable):
    conf = Dict(
        default_value=dict(
            broker_url="amqp://localhost",
            result_backend="rpc://",
            task_serializer="json",
            result_serializer="json",
            accept_content=["json"],
            broker_connection_retry_on_startup=True,
        ),
        help="Configuration for Celery app.",
    ).tag(config=True)

    worker_kwargs = Dict(
        default_value={}, help="Keyword arguments to pass to celery Worker instance."
    ).tag(config=True)

    app: Celery
    _session_maker: Union[scoped_session, None] = None

    def __init__(self, config_file: Union[str, None] = None, **kwargs):
        super().__init__(**kwargs)
        if not self.config:
            self.config_file = config_file
            if config_file is None:
                raise ValueError("Neither config nor config_path were passed to CeleryApp!")

            from grader_service.main import GraderService

            service = GraderService.instance()
            # config might not be loaded if the celery app was not initialized by the service (e.g. in a worker)
            if len(service.loaded_config_files) == 0:
                self.log.info(f"Loading GraderService config from {config_file}")
                service.load_config_file(self.config_file)
                service.set_config()
                self.update_config(service.config)

        from grader_service.autograding.celery.tasks import app

        self.app = app  # update module level celery app from tasks.py
        self.app.conf.update(self.conf)

    @property
    def sessionmaker(self) -> scoped_session:
        if self._session_maker is None:
            self.log.info("Instantiating database connection")
            from grader_service.main import GraderService, get_session_maker

            db_url = GraderService.instance().db_url
            self._session_maker = get_session_maker(db_url)
        return self._session_maker



"""Utilities for hooking up oauth2 to JupyterHub's database

implements https://oauthlib.readthedocs.io/en/latest/oauth2/server.html
"""

import oauthlib.common
from oauthlib import uri_validate
from oauthlib.oauth2 import RequestValidator, WebApplicationServer
from oauthlib.oauth2.rfc6749.grant_types import authorization_code, base
from tornado.log import app_log

from grader_service.orm.api_token import APIToken
from grader_service.orm.oauthclient import OAuthClient
from grader_service.orm.oauthcode import OAuthCode
from grader_service.utils import compare_token, hash_token

# patch absolute-uri check
# because we want to allow relative uri oauth
# for internal services


def is_absolute_uri(uri):
    if uri.startswith("/"):
        return True
    return uri_validate.is_absolute_uri(uri)


authorization_code.is_absolute_uri = is_absolute_uri
base.is_absolute_uri = is_absolute_uri


class JupyterHubRequestValidator(RequestValidator):
    def __init__(self, db):
        self.db = db
        super().__init__()

    def authenticate_client(self, request, *args, **kwargs):
        """Authenticate client through means outside the OAuth 2 spec.
        Means of authentication is negotiated beforehand and may for example
        be `HTTP Basic Authentication Scheme`_ which utilizes the Authorization
        header.
        Headers may be accesses through request.headers and parameters found in
        both body and query can be obtained by direct attribute access, i.e.
        request.client_id for client_id in the URL query.
        :param request: oauthlib.common.Request
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
            - Resource Owner Password Credentials Grant (may be disabled)
            - Client Credentials Grant
            - Refresh Token Grant
        .. _`HTTP Basic Authentication Scheme`: https://tools.ietf.org/html/rfc1945#section-11.1
        """
        app_log.debug("authenticate_client %s", request)
        client_id = request.client_id
        client_secret = request.client_secret
        oauth_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if oauth_client is None:
            return False
        if not client_secret or not oauth_client.secret:
            # disallow authentication with no secret
            return False
        if not compare_token(oauth_client.secret, client_secret):
            app_log.warning("Client secret mismatch for %s", client_id)
            return False

        request.client = oauth_client
        return True

    def authenticate_client_id(self, client_id, request, *args, **kwargs):
        """Ensure client_id belong to a non-confidential client.
        A non-confidential client is one that is not required to authenticate
        through other means, such as using HTTP Basic.
        Note, while not strictly necessary it can often be very convenient
        to set request.client to the client object associated with the
        given client_id.
        :param request: oauthlib.common.Request
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
        """
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if orm_client is None:
            app_log.warning("No such oauth client %s", client_id)
            return False
        request.client = orm_client
        return True

    def confirm_redirect_uri(
        self, client_id, code, redirect_uri, client: OAuthClient, *args, **kwargs
    ):
        """Ensure that the authorization process represented by this authorization
        code began with this 'redirect_uri'.
        If the client specifies a redirect_uri when obtaining code then that
        redirect URI must be bound to the code and verified equal in this
        method, according to RFC 6749 section 4.1.3.  Do not compare against
        the client's allowed redirect URIs, but against the URI used when the
        code was saved.
        :param client_id: Unicode client identifier
        :param code: Unicode authorization_code.
        :param redirect_uri: Unicode absolute URI
        :param client: OAuthClient object set by you, see authenticate_client.
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant (during token request)
        """
        # TODO: record redirect_uri used during oauth
        # if we ever support multiple destinations
        app_log.debug(
            "confirm_redirect_uri: client_id=%s, redirect_uri=%s", client_id, redirect_uri
        )
        if redirect_uri == client.redirect_uri:
            return True
        else:
            app_log.warning("Redirect uri %s != %s", redirect_uri, client.redirect_uri)
            return False

    def get_default_redirect_uri(self, client_id, request, *args, **kwargs):
        """Get the default redirect URI for the client.
        :param client_id: Unicode client identifier
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: The default redirect URI for the client
        Method is used by:
            - Authorization Code Grant
            - Implicit Grant
        """
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if orm_client is None:
            raise KeyError(client_id)
        return orm_client.redirect_uri

    def get_original_scopes(self, refresh_token, request, *args, **kwargs):
        """Get the list of scopes associated with the refresh token.
        :param refresh_token: Unicode refresh token
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: List of scopes.
        Method is used by:
            - Refresh token grant
        """
        raise NotImplementedError()

    def is_within_original_scope(self, request_scopes, refresh_token, request, *args, **kwargs):
        """Check if requested scopes are within a scope of the refresh token.
        When access tokens are refreshed the scope of the new token
        needs to be within the scope of the original token. This is
        ensured by checking that all requested scopes strings are on
        the list returned by the get_original_scopes. If this check
        fails, is_within_original_scope is called. The method can be
        used in situations where returning all valid scopes from the
        get_original_scopes is not practical.
        :param request_scopes: A list of scopes that were requested by client
        :param refresh_token: Unicode refresh_token
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Refresh token grant
        """
        raise NotImplementedError()

    def invalidate_authorization_code(self, client_id, code, request, *args, **kwargs):
        """Invalidate an authorization code after use.
        :param client_id: Unicode client identifier
        :param code: The authorization code grant (request.code).
        :param request: The HTTP Request (oauthlib.common.Request)
        Method is used by:
            - Authorization Code Grant
        """
        app_log.debug("Deleting oauth code %s... for %s", code[:3], client_id)
        orm_code = self.db.query(OAuthCode).filter_by(code=code).first()
        if orm_code is not None:
            self.db.delete(orm_code)
            self.db.commit()

    def revoke_token(self, token, token_type_hint, request, *args, **kwargs):
        """Revoke an access or refresh token.
        :param token: The token string.
        :param token_type_hint: access_token or refresh_token.
        :param request: The HTTP Request (oauthlib.common.Request)
        Method is used by:
            - Revocation Endpoint
        """
        app_log.debug("Revoking %s %s", token_type_hint, token[:3] + "...")
        raise NotImplementedError("Subclasses must implement this method.")

    def save_authorization_code(self, client_id, code, request, *args, **kwargs):
        """Persist the authorization_code.
        The code should at minimum be stored with:
            - the client_id (client_id)
            - the redirect URI used (request.redirect_uri)
            - a resource owner / user (request.user)
            - the authorized scopes (request.scopes)
            - the client state, if given (code.get('state'))
        The 'code' argument is actually a dictionary, containing at least a
        'code' key with the actual authorization code:
            {'code': 'sdf345jsdf0934f'}
        It may also have a 'state' key containing a nonce for the client, if it
        chose to send one.  That value should be saved and used in
        'validate_code'.
        It may also have a 'claims' parameter which, when present, will be a dict
        deserialized from JSON as described at
        http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter
        This value should be saved in this method and used again in 'validate_code'.
        :param client_id: Unicode client identifier
        :param code: A dict of the authorization code grant and, optionally, state.
        :param request: The HTTP Request (oauthlib.common.Request)
        Method is used by:
            - Authorization Code Grant
        """
        log_code = code.get("code", "undefined")[:3] + "..."
        app_log.debug("Saving authorization code %s, %s, %s, %s", client_id, log_code, args, kwargs)
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if orm_client is None:
            raise ValueError("No such client: %s" % client_id)

        orm_code = OAuthCode(
            code=code["code"],
            # oauth has 5 minutes to complete
            expires_at=int(OAuthCode.now() + 300),
            scopes=list(request.scopes),
            redirect_uri=orm_client.redirect_uri,
            session_id=request.session_id,
            user_id=request.user.id,
        )
        self.db.add(orm_code)
        orm_code.client = orm_client
        self.db.commit()

    def get_authorization_code_scopes(self, client_id, code, redirect_uri, request):
        """Extracts scopes from saved authorization code.
        The scopes returned by this method is used to route token requests
        based on scopes passed to Authorization Code requests.
        With that the token endpoint knows when to include OpenIDConnect
        id_token in token response only based on authorization code scopes.
        Only code param should be sufficient to retrieve grant code from
        any storage you are using, `client_id` and `redirect_uri` can gave a
        blank value `""` don't forget to check it before using those values
        in a select query if a database is used.
        :param client_id: Unicode client identifier
        :param code: Unicode authorization code grant
        :param redirect_uri: Unicode absolute URI
        :return: A list of scope
        Method is used by:
            - Authorization Token Grant Dispatcher
        """
        raise NotImplementedError("TODO")

    def save_token(self, token, request, *args, **kwargs):
        """Persist the token with a token type specific method.
        Currently, only save_bearer_token is supported.
        """
        return self.save_bearer_token(token, request, *args, **kwargs)

    def save_bearer_token(self, token, request, *args, **kwargs):
        """Persist the Bearer token.
        The Bearer token should at minimum be associated with:
            - a client and it's client_id, if available
            - a resource owner / user (request.user)
            - authorized scopes (request.scopes)
            - an expiration time
            - a refresh token, if issued
            - a claims document, if present in request.claims
        The Bearer token dict may hold a number of items::
            {
                'token_type': 'Bearer',
                'access_token': 'askfjh234as9sd8',
                'expires_in': 3600,
                'scope': 'string of space separated authorized scopes',
                'refresh_token': '23sdf876234',  # if issued
                'state': 'given_by_client',  # if supplied by client
            }
        Note that while "scope" is a string-separated list of authorized scopes,
        the original list is still available in request.scopes.
        The token dict is passed as a reference so any changes made to the dictionary
        will go back to the user.  If additional information must return to the client
        user, and it is only possible to get this information after writing the token
        to storage, it should be added to the token dictionary.  If the token
        dictionary must be modified but the changes should not go back to the user,
        a copy of the dictionary must be made before making the changes.
        Also note that if an Authorization Code grant request included a valid claims
        parameter (for OpenID Connect) then the request.claims property will contain
        the claims dict, which should be saved for later use when generating the
        id_token and/or UserInfo response content.
        :param token: A Bearer token dict
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: The default redirect URI for the client
        Method is used by all core grant types issuing Bearer tokens:
            - Authorization Code Grant
            - Implicit Grant
            - Resource Owner Password Credentials Grant (might not associate a client)
            - Client Credentials grant
        """
        log_token = {}
        log_token.update(token)
        # redact sensitive keys in log
        for key in ("access_token", "refresh_token", "state"):
            if key in token:
                value = token[key]
                if isinstance(value, str):
                    log_token[key] = "REDACTED"
        app_log.debug("Saving bearer token %s", log_token)

        if request.user is None:
            raise ValueError("No user for access token: %s" % request.user)
        client = self.db.query(OAuthClient).filter_by(identifier=request.client.client_id).first()
        # FIXME: support refresh tokens
        # These should be in a new table
        token.pop("refresh_token", None)

        # APIToken.new commits the token to the db
        # TODO the expires in value is hard coded
        APIToken.new(
            oauth_client=client,
            # expires_in=token['expires_in']
            expires_in=1209600,  # 14 days
            scopes=request.scopes,
            token=token["access_token"],
            session_id=request.session_id,
            user=request.user,
        )
        return client.redirect_uri

    def validate_bearer_token(self, token, scopes, request):
        """Ensure the Bearer token is valid and authorized access to scopes.
        :param token: A string of random characters.
        :param scopes: A list of scopes associated with the protected resource.
        :param request: The HTTP Request (oauthlib.common.Request)
        A key to OAuth 2 security and restricting impact of leaked tokens is
        the short expiration time of tokens, *always ensure the token has not
        expired!*.
        Two different approaches to scope validation:
            1) all(scopes). The token must be authorized access to all scopes
                            associated with the resource. For example, the
                            token has access to ``read-only`` and ``images``,
                            thus the client can view images but not upload new.
                            Allows for fine grained access control through
                            combining various scopes.
            2) any(scopes). The token must be authorized access to one of the
                            scopes associated with the resource. For example,
                            token has access to ``read-only-images``.
                            Allows for fine grained, although arguably less
                            convenient, access control.
        A powerful way to use scopes would mimic UNIX ACLs and see a scope
        as a group with certain privileges. For a restful API these might
        map to HTTP verbs instead of read, write and execute.
        Note, the request.user attribute can be set to the resource owner
        associated with this token. Similarly the request.client and
        request.scopes attribute can be set to associated client object
        and authorized scopes. If you then use a decorator such as the
        one provided for django these attributes will be made available
        in all protected views as keyword arguments.
        :param token: Unicode Bearer token
        :param scopes: List of scopes (defined by you)
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is indirectly used by all core Bearer token issuing grant types:
            - Authorization Code Grant
            - Implicit Grant
            - Resource Owner Password Credentials Grant
            - Client Credentials Grant
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def validate_client_id(self, client_id, request, *args, **kwargs):
        """Ensure client_id belong to a valid and active client.
        Note, while not strictly necessary it can often be very convenient
        to set request.client to the client object associated with the
        given client_id.
        :param request: oauthlib.common.Request
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
            - Implicit Grant
        """
        app_log.debug("Validating client id %s", client_id)
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if orm_client is None:
            return False
        if not orm_client.secret:
            return False
        request.client = orm_client
        return True

    def validate_code(self, client_id, code, client, request, *args, **kwargs):
        """Verify that the authorization_code is valid and assigned to the given
        client.
        Before returning true, set the following based on the information stored
        with the code in 'save_authorization_code':
            - request.user
            - request.state (if given)
            - request.scopes
            - request.claims (if given)
        OBS! The request.user attribute should be set to the resource owner
        associated with this authorization code. Similarly request.scopes
        must also be set.
        The request.claims property, if it was given, should assigned a dict.
        :param client_id: Unicode client identifier
        :param code: Unicode authorization code
        :param client: Client object set by you, see authenticate_client.
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
        """
        orm_code = OAuthCode.find(self.db, code=code)
        if orm_code is None:
            app_log.debug("No such code: %s", code)
            return False
        if orm_code.client_id != client_id:
            app_log.debug("OAuth code client id mismatch: %s != %s", client_id, orm_code.client_id)
            return False
        request.user = orm_code.user
        request.session_id = orm_code.session_id
        request.scopes = orm_code.scopes
        return True

    def validate_grant_type(self, client_id, grant_type, client, request, *args, **kwargs):
        """Ensure client is authorized to use the grant_type requested.
        :param client_id: Unicode client identifier
        :param grant_type: Unicode grant type, i.e. authorization_code, password.
        :param client: Client object set by you, see authenticate_client.
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
            - Resource Owner Password Credentials Grant
            - Client Credentials Grant
            - Refresh Token Grant
        """
        return grant_type == "authorization_code"

    def validate_redirect_uri(self, client_id, redirect_uri, request, *args, **kwargs):
        """Ensure client is authorized to redirect to the redirect_uri requested.
        All clients should register the absolute URIs of all URIs they intend
        to redirect to. The registration is outside of the scope of oauthlib.
        :param client_id: Unicode client identifier
        :param redirect_uri: Unicode absolute URI
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
            - Implicit Grant
        """
        app_log.debug(
            "validate_redirect_uri: client_id=%s, redirect_uri=%s", client_id, redirect_uri
        )
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if orm_client is None:
            app_log.warning("No such oauth client %s", client_id)
            return False
        if redirect_uri == orm_client.redirect_uri:
            return True
        else:
            app_log.warning("Redirect uri %s != %s", redirect_uri, orm_client.redirect_uri)
            return False

    def validate_refresh_token(self, refresh_token, client, request, *args, **kwargs):
        """Ensure the Bearer token is valid and authorized access to scopes.
        OBS! The request.user attribute should be set to the resource owner
        associated with this refresh token.
        :param refresh_token: Unicode refresh token
        :param client: Client object set by you, see authenticate_client.
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant (indirectly by issuing refresh tokens)
            - Resource Owner Password Credentials Grant (also indirectly)
            - Refresh Token Grant
        """
        return False
        raise NotImplementedError("Subclasses must implement this method.")

    def validate_response_type(self, client_id, response_type, client, request, *args, **kwargs):
        """Ensure client is authorized to use the response_type requested.
        :param client_id: Unicode client identifier
        :param response_type: Unicode response type, i.e. code, token.
        :param client: Client object set by you, see authenticate_client.
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by:
            - Authorization Code Grant
            - Implicit Grant
        """
        # TODO
        return True

    def get_default_scopes(self, client_id: str, request: oauthlib.common.Request, *args, **kwargs):
        """
        Grader Service now can only provide identification.

        :param client_id: Unicode client identifier.
        :param request: OAuthlib request.
        :return: List of default scopes
        """
        return ["identify"]

    def validate_scopes(self, client_id, scopes, client, request, *args, **kwargs):
        """Ensure the client is authorized access to requested scopes.
        :param client_id: Unicode client identifier
        :param scopes: List of 'raw' scopes (defined by you)
        :param client: Client object set by you, see authenticate_client.
        :param request: The HTTP Request (oauthlib.common.Request)
        :rtype: True or False
        Method is used by all core grant types:
            - Authorization Code Grant
            - Implicit Grant
            - Resource Owner Password Credentials Grant
            - Client Credentials Grant
        """
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).one_or_none()
        if orm_client is None:
            app_log.warning("No such oauth client %s", client_id)
            return False

        # we always grant 'identify' scope to a valid client (JupyterHub)
        return True


class GraderOAuthServer(WebApplicationServer):
    def __init__(self, db, validator, *args, **kwargs):
        self.db = db
        super().__init__(validator, *args, **kwargs)

    def add_client(
        self, client_id, client_secret, redirect_uri, allowed_scopes=None, description=""
    ):
        """Add a client

        hash its client_secret before putting it in the database.
        """
        # Update client if it already exists, else create it
        # Sqlalchemy doesn't have a good db agnostic UPSERT,
        # so we do this manually. It's protected inside a
        # transaction, so should fail if there are multiple
        # rows with the same identifier.
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).one_or_none()
        if orm_client is None:
            orm_client = OAuthClient(identifier=client_id)
            self.db.add(orm_client)
            app_log.info(f"Creating oauth client {client_id}")
        else:
            app_log.info(f"Updating oauth client {client_id}")
        if allowed_scopes is None:
            allowed_scopes = []
        orm_client.secret = hash_token(client_secret) if client_secret else ""
        orm_client.redirect_uri = redirect_uri
        orm_client.description = description or client_id
        orm_client.allowed_scopes = list(allowed_scopes)
        self.db.commit()
        return orm_client

    def remove_client(self, client_id):
        """Remove a client by its id if it is existed."""
        orm_client = self.db.query(OAuthClient).filter_by(identifier=client_id).one_or_none()
        if orm_client is not None:
            self.db.delete(orm_client)
            self.db.commit()
            app_log.info("Removed client %s", client_id)
        else:
            app_log.warning("No such client %s", client_id)

    def fetch_by_client_id(self, client_id):
        """Find a client by its id"""
        client = self.db.query(OAuthClient).filter_by(identifier=client_id).first()
        if client and client.secret:
            return client


def make_provider(session_factory, url_prefix, login_url, **oauth_server_kwargs):
    """Make an OAuth provider"""
    db = session_factory()
    oauth_server_kwargs.update(dict(url_prefix=url_prefix, login_url=login_url))
    validator = JupyterHubRequestValidator(db)
    server = GraderOAuthServer(db, validator, **oauth_server_kwargs)
    return server



import os
import sys
import typing
from textwrap import dedent, fill

from nbconvert.filters import ansi2html, strip_ansi
from nbformat import current_nbformat
from nbformat import read as read_nb
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Integer, List, Unicode
from traitlets.config import LoggingConfigurable

from grader_service.convert import utils
from grader_service.convert.preprocessors import CheckCellMetadata, ClearOutput, Execute


class Validator(LoggingConfigurable):
    preprocessors = List([CheckCellMetadata, ClearOutput, Execute])

    indent = Unicode(
        "    ", help="A string containing whitespace that will be used to indent code and errors"
    ).tag(config=True)

    width = Integer(90, help="Maximum line width for displaying code/errors").tag(config=True)

    invert = Bool(False, help="Complain when cells pass, rather than fail.").tag(config=True)

    ignore_checksums = Bool(
        False,
        help=dedent(
            """
            Don't complain if cell checksums have changed (if they are locked
            cells) or haven't changed (if they are solution cells). Note that
            this will NOT ignore changes to cell types.
            """
        ),
    ).tag(config=True)

    type_changed_warning = Unicode(
        dedent(
            """
            THE TYPES OF {num_changed} CELL(S) HAVE CHANGED!
            This might mean that even though the tests are passing
            now, they won't pass when your assignment is graded.
            """
        ).strip()
        + "\n",
        help="Warning to display when a cell's type has changed.",
    ).tag(config=True)

    changed_warning = Unicode(
        dedent(
            """
            THE CONTENTS OF {num_changed} TEST CELL(S) HAVE CHANGED!
            This might mean that even though the tests are passing
            now, they won't pass when your assignment is graded.
            """
        ).strip()
        + "\n",
        help="Warning to display when a cell has changed.",
    ).tag(config=True)

    failed_warning = Unicode(
        dedent(
            """
            VALIDATION FAILED ON {num_failed} CELL(S)! If you submit
            your assignment as it is, you WILL NOT receive full
            credit.
            """
        ).strip()
        + "\n",
        help="Warning to display when a cell fails.",
    ).tag(config=True)

    passed_warning = Unicode(
        dedent(
            """
            NOTEBOOK PASSED ON {num_passed} CELL(S)!
            """
        ).strip()
        + "\n",
        help="Warning to display when a cell passes (when invert=True)",
    ).tag(config=True)

    validate_all = Bool(False, help="Validate all cells, not just the graded tests cells.").tag(
        config=True
    )

    stream = sys.stdout

    def _indent(self, val: str) -> str:
        lines = val.split("\n")
        new_lines = []
        for line in lines:
            new_line = self.indent + strip_ansi(line)
            if len(new_line) > (self.width - 3):
                new_line = new_line[: (self.width - 3)] + "..."
            new_lines.append(new_line)
        return "\n".join(new_lines)

    def _extract_error(self, cell: NotebookNode) -> str:
        errors = []

        # possibilities:
        # 1. the cell returned an error in cell.outputs
        # 2. grade cell returned score < max_points (i.e. partial credit)
        # 3. the student did not provide a response
        if cell.cell_type == "code":
            for output in cell.outputs:
                if output.output_type == "error":
                    errors.append("\n".join(output.traceback))
                if output.output_type == "stream" and output.name == "stderr":
                    errors.append(output.text)

            if len(errors) == 0:
                if utils.is_grade(cell):
                    score, max_score = utils.determine_grade(cell, self.log)
                    if score < max_score:
                        errors.append("Partial credit; passed some but not all of the tests")
                else:
                    errors.append("You did not provide a response.")

        else:
            errors.append("You did not provide a response.")

        return "\n".join(errors)

    def _print_type_changed(self, old_type: str, new_type: str, source: str) -> None:
        self.stream.write("\n" + "=" * self.width + "\n")
        self.stream.write(
            "The following {} cell has changed to a {} cell:\n\n".format(old_type, new_type)
        )
        self.stream.write(self._indent(source) + "\n\n")

    def _print_changed(self, source):
        self.stream.write("\n" + "=" * self.width + "\n")
        self.stream.write("The following cell has changed:\n\n")
        self.stream.write(self._indent(source) + "\n\n")

    def _print_error(self, source: str, error: str) -> None:
        self.stream.write("\n" + "=" * self.width + "\n")
        self.stream.write("The following cell failed:\n\n")
        self.stream.write(self._indent(source) + "\n\n")
        self.stream.write("The error was:\n\n")
        self.stream.write(self._indent(error) + "\n\n")

    def _print_pass(self, source):
        self.stream.write("\n" + "=" * self.width + "\n")
        self.stream.write("The following cell passed:\n\n")
        self.stream.write(self._indent(source) + "\n\n")

    def _print_num_type_changed(self, num_changed: int) -> None:
        if num_changed == 0:
            return

        else:
            self.stream.write(
                fill(self.type_changed_warning.format(num_changed=num_changed), width=self.width)
            )

    def _print_num_changed(self, num_changed: int) -> None:
        if num_changed == 0:
            return

        else:
            self.stream.write(
                fill(self.changed_warning.format(num_changed=num_changed), width=self.width)
            )

    def _print_num_failed(self, num_failed: int) -> None:
        if num_failed == 0:
            self.stream.write("Success! Your notebook passes all the tests.\n")

        else:
            self.stream.write(
                fill(self.failed_warning.format(num_failed=num_failed), width=self.width)
            )

    def _print_num_passed(self, num_passed):
        if num_passed == 0:
            self.stream.write("Success! The notebook does not pass any tests.\n")

        else:
            self.stream.write(
                fill(self.passed_warning.format(num_passed=num_passed), width=self.width)
            )

    def _get_type_changed_cells(self, nb: NotebookNode) -> typing.List[NotebookNode]:
        changed = []

        for cell in nb.cells:
            if not (utils.is_grade(cell) or utils.is_solution(cell) or utils.is_locked(cell)):
                continue
            if "cell_type" not in cell.metadata.nbgrader:
                continue

            new_type = cell.metadata.nbgrader.cell_type
            old_type = cell.cell_type
            if new_type and (old_type != new_type):
                changed.append(cell)

        return changed

    def _get_changed_cells(self, nb: NotebookNode) -> typing.List:
        changed = []
        for cell in nb.cells:
            if not (utils.is_grade(cell) or utils.is_locked(cell)):
                continue

            # if we're ignoring checksums, then remove the checksum from the
            # cell metadata
            if self.ignore_checksums and "checksum" in cell.metadata.nbgrader:
                del cell.metadata.nbgrader["checksum"]

            # verify checksums of cells
            if utils.is_locked(cell) and "checksum" in cell.metadata.nbgrader:
                old_checksum = cell.metadata.nbgrader["checksum"]
                new_checksum = utils.compute_checksum(cell)
                if old_checksum != new_checksum:
                    changed.append(cell)

        return changed

    def _get_failed_cells(self, nb: NotebookNode) -> typing.List[NotebookNode]:
        failed = []
        for cell in nb.cells:
            if not (self.validate_all or utils.is_grade(cell) or utils.is_locked(cell)):
                continue

            # if it's a grade cell, the check the grade
            if utils.is_grade(cell):
                score, max_score = utils.determine_grade(cell, self.log)

                # it's a markdown cell, so we can't do anything
                if score is None:
                    pass
                elif score < max_score:
                    failed.append(cell)
            elif self.validate_all and cell.cell_type == "code":
                for output in cell.outputs:
                    if (
                        output.output_type == "error"
                        or output.output_type == "stream"
                        and output.name == "stderr"
                    ):
                        failed.append(cell)
                        break

        return failed

    def _get_passed_cells(self, nb: NotebookNode) -> typing.List[NotebookNode]:
        passed = []
        for cell in nb.cells:
            if not (utils.is_grade(cell) or utils.is_locked(cell)):
                continue

            # if it's a grade cell, the check the grade
            if utils.is_grade(cell):
                score, max_score = utils.determine_grade(cell, self.log)

                # it's a markdown cell, so we can't do anything
                if score is None:
                    pass
                elif score == max_score:
                    passed.append(cell)

        return passed

    def _preprocess(self, nb: NotebookNode) -> NotebookNode:
        resources = {}
        with utils.setenv(NBGRADER_VALIDATING="1"):
            for preprocessor in self.preprocessors:
                # https://github.com/jupyter/nbgrader/pull/1075
                # It seemes that without the self.config passed below,
                # --ExecutePreprocessor.timeout doesn't work.  Better solution
                # requested, unknown why this is needed.
                pp = preprocessor(**self.config[preprocessor.__name__])
                nb, resources = pp.preprocess(nb, resources)
        return nb

    def validate(self, filename: str) -> typing.Dict[str, typing.List[typing.Dict[str, str]]]:
        self.log.info("Validating '{}'".format(os.path.abspath(filename)))
        basename = os.path.basename(filename)
        dirname = os.path.dirname(filename)
        with utils.chdir(dirname):
            nb = read_nb(basename, as_version=current_nbformat)

        type_changed = self._get_type_changed_cells(nb)
        if len(type_changed) > 0:
            results = {}
            results["type_changed"] = [
                {
                    "source": cell.source.strip(),
                    "old_type": cell.cell_type,
                    "new_type": cell.metadata.nbgrader.cell_type,
                }
                for cell in type_changed
            ]
            return results

        with utils.chdir(dirname):
            nb = self._preprocess(nb)
        changed = self._get_changed_cells(nb)
        passed = self._get_passed_cells(nb)
        failed = self._get_failed_cells(nb)

        results = {}

        if not self.ignore_checksums and len(changed) > 0:
            results["changed"] = [{"source": cell.source.strip()} for cell in changed]

        elif self.invert:
            if len(passed) > 0:
                results["passed"] = [{"source": cell.source.strip()} for cell in passed]

        else:
            if len(failed) > 0:
                results["failed"] = [
                    {
                        "source": cell.source.strip(),
                        "error": ansi2html(self._extract_error(cell)),
                        "raw_error": self._extract_error(cell),
                    }
                    for cell in failed
                ]

        return results

    def validate_and_print(self, filename: str) -> None:
        results = self.validate(filename)
        type_changed = results.get("type_changed", [])
        changed = results.get("changed", [])
        passed = results.get("passed", [])
        failed = results.get("failed", [])

        if len(type_changed) > 0:
            self._print_num_type_changed(len(type_changed))
            for cell in type_changed:
                self._print_type_changed(cell["old_type"], cell["new_type"], cell["source"])

        elif not self.ignore_checksums and len(changed) > 0:
            self._print_num_changed(len(changed))
            for cell in changed:
                self._print_changed(cell["source"])

        elif self.invert:
            self._print_num_passed(len(passed))
            for cell in passed:
                self._print_pass(cell["source"])

        else:
            self._print_num_failed(len(failed))
            for cell in failed:
                self._print_error(cell["source"], cell["raw_error"])



import typing

from nbformat import read as _read
from nbformat import reads as _reads
from nbformat import write as _write
from nbformat import writes as _writes
from nbformat.notebooknode import NotebookNode

from .common import BaseMetadataValidator, ValidationError
from .v1 import MetadataValidatorV1


class MetadataValidatorV2(BaseMetadataValidator):
    schema_version = 2

    def __init__(self) -> None:
        super().__init__()
        self.v1 = MetadataValidatorV1()

    def _upgrade_v1_to_v2(self, cell: NotebookNode) -> NotebookNode:
        meta = cell.metadata["nbgrader"]

        # only add cell type if the checksum has also already been set
        if "checksum" in meta and "cell_type" not in meta:
            self.log.warning("Cell does not have a stored cell type! Adding default cell type.")
            meta["cell_type"] = cell.cell_type

        meta["schema_version"] = self.schema_version

        return cell

    def upgrade_cell_metadata(self, cell: NotebookNode) -> NotebookNode:
        if "nbgrader" not in cell.metadata:
            return cell

        if "schema_version" not in cell.metadata["nbgrader"]:
            cell.metadata["nbgrader"]["schema_version"] = 0

        if cell.metadata["nbgrader"]["schema_version"] == 0:
            cell = self.v1._upgrade_v0_to_v1(cell)
            if "nbgrader" not in cell.metadata:
                return cell

        if cell.metadata["nbgrader"]["schema_version"] == 1:
            cell = self._upgrade_v1_to_v2(cell)

        self._remove_extra_keys(cell)
        return cell

    def validate_cell(self, cell: NotebookNode) -> None:
        super(MetadataValidatorV2, self).validate_cell(cell)

        if "nbgrader" not in cell.metadata:
            return

        meta = cell.metadata["nbgrader"]
        grade = meta["grade"]
        solution = meta["solution"]
        locked = meta["locked"]

        # check if the cell type has changed
        if "cell_type" in meta:
            if meta["cell_type"] != cell.cell_type:
                self.log.warning(
                    "Cell type has changed from {} to {}!".format(
                        meta["cell_type"], cell.cell_type
                    ),
                    cell,
                )

        # check for a valid grade id
        if grade or solution or locked:
            if "grade_id" not in meta:
                raise ValidationError(
                    "nbgrader cell does not have a grade_id: {}".format(cell.source)
                )
            if meta["grade_id"] == "":
                raise ValidationError("grade_id is empty")

        # check for valid points
        if grade:
            if "points" not in meta:
                raise ValidationError(
                    "nbgrader cell '{}' does not have points".format(meta["grade_id"])
                )

        # check that markdown cells are grade AND solution (not either/or)
        if cell.cell_type == "markdown" and grade and not solution:
            raise ValidationError(
                "Markdown grade cell '{}' is not marked as a solution cell".format(meta["grade_id"])
            )
        if cell.cell_type == "markdown" and not grade and solution:
            raise ValidationError(
                "Markdown solution cell is not marked as a grade cell: {}".format(cell.source)
            )

    def validate_nb(self, nb: NotebookNode) -> None:
        super(MetadataValidatorV2, self).validate_nb(nb)

        ids = set([])
        for cell in nb.cells:
            if "nbgrader" not in cell.metadata:
                continue

            grade = cell.metadata["nbgrader"]["grade"]
            solution = cell.metadata["nbgrader"]["solution"]
            locked = cell.metadata["nbgrader"]["locked"]

            if not grade and not solution and not locked:
                continue

            grade_id = cell.metadata["nbgrader"]["grade_id"]
            if grade_id in ids:
                raise ValidationError("Duplicate grade id: {}".format(grade_id))
            ids.add(grade_id)


def read_v2(source: typing.TextIO, as_version: int, **kwargs: typing.Any) -> NotebookNode:
    nb = _read(source, as_version, **kwargs)
    MetadataValidatorV2().validate_nb(nb)
    return nb


def write_v2(nb: NotebookNode, fp: typing.TextIO, **kwargs: typing.Any) -> None:
    MetadataValidatorV2().validate_nb(nb)
    _write(nb, fp, **kwargs)


def reads_v2(source: str, as_version: int, **kwargs: typing.Any) -> NotebookNode:
    nb = _reads(source, as_version, **kwargs)
    MetadataValidatorV2().validate_nb(nb)
    return nb


def writes_v2(nb: NotebookNode, **kwargs: typing.Any) -> None:
    MetadataValidatorV2().validate_nb(nb)
    _writes(nb, **kwargs)



import json
import logging
import os
from functools import wraps
from typing import Any, Optional, Union

from grader_service.convert.gradebook.models import (
    Comment,
    Grade,
    GradeBookModel,
    GradeCell,
    Notebook,
    SolutionCell,
    SourceCell,
    TaskCell,
)


class InvalidEntry(ValueError):
    pass


class MissingEntry(ValueError):
    pass


def write_access(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            self: "Gradebook" = args[0]
            if not isinstance(self, Gradebook):
                raise RuntimeError(
                    "This decorator can only be used on methods inside Gradebook instances!"
                )
        except IndexError:
            raise RuntimeError(
                "This decorator can only be used on methods inside Gradebook instances!"
            )
        self.dirty = True
        result = fn(*args, **kwargs)
        if not self.is_in_context:
            self.write_model()
            self.dirty = False
        return result

    return wrapper


# TODO: add decorator to functions that sets dirty flag for methods and checks on __enter__
class Gradebook:
    """
    The gradebook object to interface with the JSON output file of a conversion.
    Should only be used as a context manager when changing the data.
    """

    def __init__(self, dest_json: str, log: logging.Logger = None) -> None:
        if log is None:
            from traitlets import log as l

            self.log = l.get_logger()
        else:
            self.log = log

        self.json_file: str = dest_json
        if os.path.isfile(self.json_file):
            with open(self.json_file, "r") as f:
                data = f.read()
                self.log.info(f"Reading {len(data)} bytes from {self.json_file}")
                self.data: dict = json.loads(data)
        else:
            os.makedirs(os.path.dirname(self.json_file), exist_ok=True)
            self.data: dict = {"notebooks": dict()}
            with open(self.json_file, "w") as f:
                json.dump(self.data, f)
        self.model: GradeBookModel = GradeBookModel.from_dict(self.data)

        self.in_context: int = 0
        self.dirty: bool = False

    def __enter__(self) -> "Gradebook":
        self.in_context += 1
        return self

    def __exit__(
        self, exc_type: Optional[Any], exc_value: Optional[Any], traceback: Optional[Any]
    ) -> None:
        self.in_context -= 1
        if self.dirty and exc_type is None:
            self.write_model()
            self.dirty = False

    @property
    def is_in_context(self) -> bool:
        return self.in_context > 0

    def write_model(self):
        """Writes JSON string to a JSON file"""
        with open(self.json_file, "w") as f:
            json_str = json.dumps(self.model.to_dict())
            self.log.info(f"Writing {len(json_str.encode('utf-8'))} bytes to {self.json_file}")
            f.write(json_str)

    # Notebooks
    @write_access
    def add_notebook(self, name: str, **kwargs: dict) -> Notebook:
        """
        Add a new notebook to the :class:`~GradeBookModel`.
        Parameters
        :param name:  the name of the new notebook
        :param kwargs: additional keyword arguments for the :class:`~Notebook` object
        :return: notebook : :class:`~Notebook`
        """
        kwargs = {k: v for k, v in kwargs.items() if k in set(Notebook.empty_dict().keys())}
        kwargs["name"] = name
        self.model.notebooks[name] = Notebook.from_dict(kwargs)
        return self.model.notebooks[name]

    def find_notebook(self, name: str) -> Notebook:
        """
        Find a particular notebook.
        :param name: the name of the notebook
        :return: notebook : :class:`~Notebook`
        :raises: MissingEntry if `name` is not a key in the self.model.notebooks
        """
        try:
            return self.model.notebooks[name]
        except KeyError:
            raise MissingEntry(name) from None

    @write_access
    def update_or_create_notebook(self, name: str, **kwargs) -> Notebook:
        """
        Update an existing notebook, or create it if it doesn't exist.
        :param name: the name of the notebook
        :param kwargs: additional keyword arguments for the :class:`~Notebook` object
        :return: notebook : :class:`~Notebook`
        """
        raise NotImplementedError()

    @write_access
    def remove_notebook(self, name: str):
        """
        Deletes an existing notebook from the gradebook.
        :param name: the name of the notebook to delete
        :return: None
        """
        self.model.notebooks.pop(name, None)

    # Grade cells

    @write_access
    def add_grade_cell(self, name: str, notebook: str, **kwargs: dict) -> GradeCell:
        """
        Add a new grade cell to an existing notebook.
        :param name: the name of the new grade cell
        :param notebook: the name of an existing notebook
        :param kwargs: additional keyword arguments for :class:`~GradeCell`
        :return: grade_cell
        """
        kwargs = {k: v for k, v in kwargs.items() if k in set(GradeCell.empty_dict().keys())}
        kwargs["name"] = name
        nb = self.find_notebook(notebook)
        nb.grade_cells_dict[name] = GradeCell.from_dict(kwargs)
        return nb.grade_cells_dict[name]

    def find_grade_cell(self, name: str, notebook: str) -> GradeCell:
        """
        Find a grade cell in a particular notebook.
        :param name: the name of the grade cell
        :param notebook: the name of the notebook
        :return: grade_cell
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.grade_cells_dict[name]
        except KeyError:
            raise MissingEntry()

    def find_graded_cell(self, name: str, notebook: str) -> Union[GradeCell, TaskCell]:
        """
        Find a graded cell in a particular notebook. This can be either a GradeCell or a TaskCell.
        :param name: the name of the grade cell
        :param notebook: the name of the notebook
        :return: grade_cell
        """
        nb = self.find_notebook(notebook)
        try:
            return [x for x in nb.graded_cells if x.name == name][0]
        except IndexError:
            raise MissingEntry()

    @write_access
    def update_or_create_grade_cell(self, name: str, notebook: str, **kwargs: dict) -> GradeCell:
        """
        Update an existing grade cell in a notebook, or create the grade cell if it does not exist.
        :param name: the name of the grade cell
        :param notebook: the name of the notebook
        :param kwargs: additional keyword arguments for :class:`~GradeCell`
        :return: grade_cell
        """
        try:
            grade_cell = self.find_grade_cell(name, notebook)
        except MissingEntry:
            grade_cell = self.add_grade_cell(name, notebook, **kwargs)
        else:
            for attr in kwargs:
                setattr(grade_cell, attr, kwargs[attr])
        return grade_cell

    # Solution cells

    @write_access
    def add_solution_cell(self, name: str, notebook: str, **kwargs: dict) -> SolutionCell:
        """
        Add a new solution cell to an existing notebook.
        :param name: the name of the new solution cell
        :param notebook: the name of an existing notebook
        :param kwargs: additional keyword arguments for :class:`~SolutionCell`
        :return: solution_cell : :class:`~SolutionCell`
        """
        kwargs = {k: v for k, v in kwargs.items() if k in set(SolutionCell.empty_dict().keys())}
        kwargs["name"] = name
        nb = self.find_notebook(notebook)
        nb.solution_cells_dict[name] = SolutionCell.from_dict(kwargs)
        return nb.solution_cells_dict[name]

    def find_solution_cell(self, name: str, notebook: str) -> SolutionCell:
        """
        Find a solution cell in a particular notebook.
        :param name: the name of the solution cell
        :param notebook: the name of the notebook
        :return: solution_cell : :class:`~SolutionCell`
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.solution_cells_dict[name]
        except KeyError:
            raise MissingEntry()

    @write_access
    def update_or_create_solution_cell(
        self, name: str, notebook: str, **kwargs: dict
    ) -> SolutionCell:
        """
        Update an existing solution cell in a notebook, or create the solution cell if it does not exist.
        :param name: the name of the solution cell
        :param notebook: the name of the notebook
        :param kwargs: additional keyword arguments for :class:`~SolutionCell`
        :return: solution_cell
        """
        try:
            solution_cell = self.find_solution_cell(name, notebook)
        except MissingEntry:
            solution_cell = self.add_solution_cell(name, notebook, **kwargs)
        else:
            for attr in kwargs:
                setattr(solution_cell, attr, kwargs[attr])
        return solution_cell

    # Task cells

    @write_access
    def add_task_cell(self, name: str, notebook: str, **kwargs: dict) -> TaskCell:
        """
        Add a new task cell to an existing notebook.
        :param name: the name of the new task cell
        :param notebook: the name of an existing notebook
        :param kwargs: additional keyword arguments for :class:`~TaskCell`
        :return: task_cell
        """
        kwargs = {k: v for k, v in kwargs.items() if k in set(TaskCell.empty_dict().keys())}
        kwargs["name"] = name
        nb = self.find_notebook(notebook)
        nb.task_cells_dict[name] = TaskCell.from_dict(kwargs)
        return nb.task_cells_dict[name]

    def find_task_cell(self, name: str, notebook: str) -> TaskCell:
        """
        Find a task cell in a particular notebook.
        :param name: the name of the task cell
        :param notebook: the name of the notebook
        :return: task_cell : :class:`~TaskCell`
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.task_cells_dict[name]
        except KeyError:
            raise MissingEntry()

    @write_access
    def update_or_create_task_cell(self, name: str, notebook: str, **kwargs) -> TaskCell:
        """
        Update an existing task cell in a notebook, or create the task cell if it does not exist.
        :param name: the name of the task cell
        :param notebook: the name of the notebook
        :param kwargs: additional keyword arguments for :class:`~TaskCell`
        :return: task_cell : :class:`~TaskCell`
        """
        try:
            task_cell = self.find_task_cell(name, notebook)
        except MissingEntry:
            task_cell = self.add_task_cell(name, notebook, **kwargs)
        else:
            for attr in kwargs:
                setattr(task_cell, attr, kwargs[attr])
        return task_cell

    # Source cell

    @write_access
    def add_source_cell(self, name: str, notebook: str, **kwargs: dict) -> SourceCell:
        """
        Add a new source cell to an existing notebook.
        :param name: the name of the new source cell
        :param notebook: the name of an existing notebook
        :param kwargs: additional keyword arguments for :class:`~SourceCell`
        :return: source_cell : :class:`~SourceCell`
        """
        kwargs = {k: v for k, v in kwargs.items() if k in set(SourceCell.empty_dict().keys())}
        kwargs["name"] = name
        nb = self.find_notebook(notebook)
        nb.source_cells_dict[name] = SourceCell.from_dict(kwargs)
        return nb.source_cells_dict[name]

    def find_source_cell(self, name: str, notebook: str) -> SourceCell:
        """
        Find a source cell in a particular notebook.
        :param name: the name of the source cell
        :param notebook: the name of the notebook
        :return: source_cell : :class:`~SourceCell`
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.source_cells_dict[name]
        except KeyError:
            raise MissingEntry()

    @write_access
    def update_or_create_source_cell(self, name: str, notebook: str, **kwargs: dict) -> SourceCell:
        """
        Update an existing source cell in a notebook of an assignment, or
        create the source cell if it does not exist.
        :param name: the name of the source cell
        :param notebook: the name of the notebook
        :param kwargs: additional keyword arguments for :class:`~SourceCell`
        :return: source_cell
        """
        try:
            source_cell = self.find_source_cell(name, notebook)
        except MissingEntry:
            source_cell = self.add_source_cell(name, notebook, **kwargs)
        else:
            for attr in kwargs:
                setattr(source_cell, attr, kwargs[attr])
        return source_cell

    # Grades

    @write_access
    def add_grade(self, grade_cell: str, notebook: str, grade: Grade) -> Grade:
        """
        Add a grade to a notebook.
        :param grade_cell: the name of a grade or task cell
        :param notebook: the name of a notebook
        :param grade: the grade to add
        :return: grade
        """
        nb = self.find_notebook(notebook)

        grade_c: GradeCell = nb.grade_cells_dict.get(grade_cell, None)
        task_c: TaskCell = nb.task_cells_dict.get(grade_cell, None)
        solution_c: SolutionCell = nb.solution_cells_dict.get(grade_cell, None)
        if all([x is None for x in [grade_c, task_c, solution_c]]):
            raise MissingEntry()
        grade.id = grade_cell
        grade.notebook_id = notebook
        grade.cell_id = grade_cell
        if grade_c is not None:
            grade.max_score_gradecell = grade_c.max_score
            grade_c.grade_id = grade.id
        if task_c is not None:
            grade.max_score_taskcell = task_c.max_score
            task_c.grade_id = grade.id
        if solution_c is not None:
            solution_c.grade_id = grade.id
        nb.grades_dict[grade.id] = grade
        return grade

    def find_grade(self, grade_cell: str, notebook: str) -> Grade:
        """
        Find a particular grade in a notebook.
        If the grade does not exists an empty grade object is returned.
        :param grade_cell: the name of a grade or task cell
        :param notebook: the name of a notebook
        :return: grade
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.grades_dict[grade_cell]
        except KeyError:
            return Grade.from_dict(Grade.empty_dict())

    @write_access
    def add_comment(self, solution_cell: str, notebook: str, comment: Comment) -> Comment:
        """
        Add a comment to a notebook.
        :param solution_cell: the name of a solution or task cell
        :param notebook: the name of a notebook
        :param comment: the comment to add
        :return: comment
        """
        nb = self.find_notebook(notebook)

        grade_c: GradeCell = nb.grade_cells_dict.get(solution_cell, None)
        task_c: TaskCell = nb.task_cells_dict.get(solution_cell, None)
        solution_c: SolutionCell = nb.solution_cells_dict.get(solution_cell, None)
        if all([x is None for x in [solution_c, task_c, solution_c]]):
            raise MissingEntry()
        comment.id = solution_cell
        comment.notebook_id = notebook
        comment.cell_id = solution_cell
        if grade_c is not None:
            grade_c.comment_id = comment.id
        if task_c is not None:
            task_c.comment_id = comment.id
        if solution_c is not None:
            solution_c.comment_id = comment.id
        nb.comments_dict[comment.id] = comment
        return comment

    def find_comment(self, solution_cell: str, notebook: str) -> Comment:
        """
        Find a particular comment in a notebook.
        :param solution_cell: the name of a solution or task cell
        :param notebook: the name of a notebook
        :return: comment
        """
        nb = self.find_notebook(notebook)
        try:
            return nb.comments_dict[solution_cell]
        except KeyError:
            return Comment.from_dict(Comment.empty_dict())

    def get_extra_files(self):
        if self.model.extra_files is None:
            return []
        return self.model.extra_files

    @write_access
    def set_extra_files(self, extra_files):
        self.model.extra_files = extra_files



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# grader_s/grader_s/handlers
import datetime
import json
import os.path
import shutil
import subprocess
from http import HTTPStatus
from typing import List

import isodate
import pandas as pd
import tornado
from celery import chain
from sqlalchemy import label
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import func
from tornado.web import HTTPError

from grader_service.api.models.submission import Submission as SubmissionModel
from grader_service.autograding.celery.tasks import (
    autograde_task,
    generate_feedback_task,
    lti_sync_task,
)
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.handlers.base_handler import GraderBaseHandler, authorize
from grader_service.handlers.handler_utils import GitRepoType, parse_ids
from grader_service.orm.assignment import Assignment
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import AutoStatus, FeedbackStatus, ManualStatus, Submission
from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties
from grader_service.orm.takepart import Role, Scope
from grader_service.orm.user import User
from grader_service.plugins.lti import LTISyncGrades
from grader_service.registry import VersionSpecifier, register_handler

# Commit hash is used to differentiate between submissions created by instructors for students and
# normal submissions by any user.
INSTRUCTOR_SUBMISSION_COMMIT_CASH = "0" * 40


def remove_points_from_submission(submissions):
    for s in submissions:
        if s.feedback_status not in (FeedbackStatus.GENERATED, FeedbackStatus.FEEDBACK_OUTDATED):
            s.score = None
    return submissions


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/submissions\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionLectureHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/submissions.
    """

    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int):
        """Return the submissions of a specific lecture.

        Two query parameter:
        1 - filter
            latest: only get the latest submissions of users.
            best: only get the best submissions by score of users.
        2 - format:
            csv: return list as comma separated values
            json: return list as JSON

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id = parse_ids(lecture_id)
        self.validate_parameters("filter", "format")
        submission_filter = self.get_argument("filter", "best")
        if submission_filter not in ["latest", "best"]:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                reason="Filter parameter has to be either 'latest' or 'best'",
            )
        response_format = self.get_argument("format", "json")
        if response_format not in ["json", "csv"]:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST, reason="Response format can either be 'json' or 'csv'"
            )

        if submission_filter == "latest":
            subquery = (
                self.session.query(Submission.user_id, func.max(Submission.date).label("max_date"))
                .join(Assignment, Submission.assignid == Assignment.id)
                .filter(Assignment.lectid == lecture_id)
                .filter(Submission.deleted == DeleteState.active)
                .group_by(Submission.user_id, Assignment.id)
                .subquery()
            )
        else:
            subquery = (
                self.session.query(
                    Submission.user_id, func.max(Submission.score).label("max_score")
                )
                .join(Assignment, Submission.assignid == Assignment.id)
                .filter(Assignment.lectid == lecture_id)
                .filter(Submission.deleted == DeleteState.active)
                .group_by(Submission.user_id, Assignment.id)
                .subquery()
            )

        submissions_query = (
            self.session.query(
                label("username", User.name),
                label("score", Submission.score),
                label("assignment", Assignment.name),
            )
            .join(Assignment, Submission.assignid == Assignment.id)
            .join(User, Submission.user_id == User.id)
            .filter(Assignment.lectid == lecture_id)
            .filter(Submission.deleted == DeleteState.active)
        )

        if submission_filter == "latest":
            submissions = (
                submissions_query.join(
                    subquery,
                    (Submission.user_id == subquery.c.user_id)
                    & (Submission.date == subquery.c.max_date),
                )
                .order_by(Assignment.id)
                .all()
            )
        else:
            submissions = (
                submissions_query.join(
                    subquery,
                    (Submission.user_id == subquery.c.user_id)
                    & (Submission.score == subquery.c.max_score),
                )
                .order_by(Submission.id)
                .all()
            )

        df = pd.DataFrame(submissions, columns=["Username", "Score", "Assignment"])
        pivoted_df = df.pivot_table(
            values="Score", index="Username", columns="Assignment", aggfunc="first", dropna=False
        ).fillna("-")

        if response_format == "csv":
            self.set_header("Content-Type", "text/csv")
            self.write(pivoted_df.to_csv(header=True, index=True))
        else:
            self.set_header("Content-Type", "application/json")
            self.write(pivoted_df.to_json(orient="columns", force_ascii=False))


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments"
    + r"\/(?P<assignment_id>\d*)\/submissions\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions.
    """

    def on_finish(self):
        # we do not close the session we just commit because we might run
        # LocalAutogradeExecutor or LocalFeedbackExecutor in POST which
        # still need it
        pass

    def validate_assignment(self, lecture_id, assignment_id):
        """Checks if assignment is part of lecture

        Args:
            lecture_id (int): lecture id
            assignment_id (int): assignment id

        Raises:
            HTTPError: raises 404 error if no assignment in lecture is found
        """
        try:
            # Query the Assignment table to check if the assignment exists and is linked to the correct lecture
            self.session.query(Assignment).filter_by(id=assignment_id, lectid=lecture_id).one()
        except NoResultFound:
            # Raise an error if no such assignment exists for the provided lecture
            raise HTTPError(
                HTTPStatus.NOT_FOUND,
                reason=f"Assignment {assignment_id} does not exist for lecture {lecture_id}",
            )
        return True

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Return the submissions of an assignment.

        Two query parameter: latest, instructor-version.

        latest: only get the latest submissions of users.
        instructor-version: if true, get the submissions of all users in
        lecture if false, get own submissions.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters("filter", "instructor-version", "format")
        submission_filter = self.get_argument("filter", "none")
        if submission_filter not in ["none", "latest", "best"]:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                reason="Filter parameter has to be either 'none', 'latest' or 'best'",
            )
        instr_version = self.get_argument("instructor-version", None) == "true"
        response_format = self.get_argument("format", "json")
        if response_format not in ["json", "csv"]:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST, reason="Response format can either be 'json' or 'csv'"
            )

        # check required scopes for instructor version
        role: Role = self.get_role(lecture_id)
        if instr_version and role.role < Scope.tutor:
            raise HTTPError(HTTPStatus.FORBIDDEN, reason="Forbidden")
        # validate that assignment is part of lecture
        self.validate_assignment(lecture_id, assignment_id)

        # get list of submissions based on arguments
        user_id = None if instr_version else role.user_id
        if submission_filter == "latest":
            submissions = self.get_latest_submissions(assignment_id, user_id=user_id)
        elif submission_filter == "best":
            submissions = self.get_best_submissions(assignment_id, user_id=user_id)
        else:
            query = self.session.query(Submission).filter(
                Submission.assignid == assignment_id, Submission.deleted == DeleteState.active
            )
            if user_id:
                query = query.filter(Submission.user_id == user_id)
            submissions = query.order_by(Submission.id).all()

        if response_format == "csv":
            self._write_csv(submissions)
        else:
            if not instr_version:
                submissions = remove_points_from_submission(submissions)

            self.write_json(submissions)
        self.session.close()

    def _write_csv(self, submissions):
        self.set_header("Content-Type", "text/csv")
        for i, s in enumerate(submissions):
            d = s.model.to_dict()
            if i == 0:
                self.write(",".join((k for k in d.keys() if k != "logs")) + "\n")
            self.write(",".join((str(v) for k, v in d.items() if k != "logs")) + "\n")

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def post(self, lecture_id: int, assignment_id: int):
        """Create submission based on commit hash.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        try:
            commit_hash = body["commit_hash"]
        except KeyError:
            raise HTTPError(400, reason="Commit hash not found in body")

        role = self.get_role(lecture_id)
        lecture = self.get_lecture(lecture_id)
        assignment = self.get_assignment(lecture_id, assignment_id)
        if assignment.status == "complete":
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Cannot submit completed assignment!")
        if role.role == Scope.student and assignment.status != "released":
            raise HTTPError(HTTPStatus.NOT_FOUND, "Assignment not found")
        # set utc time
        submission_ts = datetime.datetime.now(datetime.timezone.utc)
        # use implicit utc time to compare with database objects
        # submission_ts = submission_ts.replace(tzinfo=None)

        score_scaling = 1.0
        if assignment.settings.deadline is not None:
            score_scaling = self.calculate_late_submission_scaling(assignment, submission_ts, role)

        if assignment.settings.max_submissions and role.role < Scope.tutor:
            submissions = assignment.submissions
            user_submissions = [s for s in submissions if s.user_id == role.user_id]
            if len(user_submissions) >= assignment.settings.max_submissions:
                raise HTTPError(
                    HTTPStatus.CONFLICT, reason="Maximum number of submissions reached!"
                )

        submission = Submission()
        submission.assignid = assignment.id
        submission.date = submission_ts
        submission.score_scaling = score_scaling

        username = body.get("username")
        if username is not None and role.role >= Scope.tutor:
            # Instructor/tutor creates a submission for a student:
            s_user = self.session.query(User).filter(User.name == username).one_or_none()
            if s_user is None:
                raise HTTPError(HTTPStatus.NOT_FOUND, reason=f"User {username} not found")
            s_user_role = self.session.get(Role, (s_user.id, lecture_id))
            if s_user_role is None:
                raise HTTPError(
                    HTTPStatus.NOT_FOUND,
                    reason=f"User {username} does not take part in this lecture",
                )
            submission.user_id = s_user.id
        else:
            # A user creates a submission for themselves.
            submission.user_id = self.user.id

            git_repo_path = self.construct_git_dir(
                repo_type=GitRepoType.USER, lecture=assignment.lecture, assignment=assignment
            )

            # If no submissions for the student exists, we cannot reference a non-existing
            # commit_hash.
            if not os.path.exists(git_repo_path):
                raise HTTPError(
                    HTTPStatus.UNPROCESSABLE_ENTITY, reason="User git repository not found"
                )
            try:
                subprocess.run(
                    ["git", "branch", "main", "--contains", commit_hash],
                    cwd=git_repo_path,
                    capture_output=True,
                )
            except subprocess.CalledProcessError:
                raise HTTPError(HTTPStatus.NOT_FOUND, reason="Commit not found")

        submission.commit_hash = commit_hash
        submission.auto_status = AutoStatus.NOT_GRADED
        submission.manual_status = ManualStatus.NOT_GRADED
        submission.feedback_status = FeedbackStatus.NOT_GENERATED

        automatic_grading = assignment.settings.autograde_type

        self.session.add(submission)
        self.session.commit()
        self.set_status(HTTPStatus.CREATED)
        self.write_json(submission)

        # If the assignment has automatic or fully automatic grading, run the necessary tasks.
        # The autograding is not performed when an instructor creates a new submission for a student.
        # Note: A specific commit hash is used to differentiate between submissions created by
        # instructors for students and normal submissions by any user.
        if (
            automatic_grading in ["auto", "full_auto"]
            and commit_hash != INSTRUCTOR_SUBMISSION_COMMIT_CASH
        ):
            submission.auto_status = AutoStatus.PENDING
            self.session.commit()
            self.set_status(HTTPStatus.ACCEPTED)

            if automatic_grading == "full_auto":
                submission.feedback_status = FeedbackStatus.GENERATING
                self.session.commit()

                # use immutable signature:
                # https://docs.celeryq.dev/en/stable/reference/celery.app.task.html#celery.app.task.Task.si
                grading_chain = chain(
                    autograde_task.si(lecture_id, assignment_id, submission.id),
                    generate_feedback_task.si(lecture_id, assignment_id, submission.id),
                    lti_sync_task.si(
                        lecture.serialize(),
                        assignment.serialize(),
                        [submission.serialize_with_user()],
                        feedback_sync=True,
                    ),
                )
            else:
                grading_chain = chain(autograde_task.si(lecture_id, assignment_id, submission.id))
            grading_chain()

        if automatic_grading == "unassisted":
            self.session.close()

    @staticmethod
    def calculate_late_submission_scaling(
        assignment: Assignment, submission_ts, role: Role
    ) -> float:
        # make submission timestamp timezone aware
        deadline = assignment.settings.deadline
        if submission_ts.tzinfo is None:
            submission_ts = submission_ts.replace(tzinfo=datetime.timezone.utc)
        if deadline.tzinfo is None:
            deadline = deadline.replace(tzinfo=datetime.timezone.utc)

        if assignment.settings.late_submission and len(assignment.settings.late_submission) > 0:
            scaling = 0.0
            if submission_ts <= deadline:
                scaling = 1.0
            else:
                for period in assignment.settings.late_submission:
                    late_submission_date = deadline + isodate.parse_duration(period.period)
                    if submission_ts < late_submission_date:
                        scaling = period.scaling
                        break
                if scaling == 0.0 and role.role < Scope.tutor:
                    raise HTTPError(
                        HTTPStatus.CONFLICT,
                        reason="Submission after last late submission period of assignment!",
                    )
        else:
            if submission_ts < deadline:
                scaling = 1.0
            else:
                if role.role < Scope.tutor:
                    raise HTTPError(
                        HTTPStatus.CONFLICT, reason="Submission after due date of assignment!"
                    )
                else:
                    scaling = 0.0
        return scaling


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/"
    + r"(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionObjectHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions
    /{submission_id}.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Returns a specific submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id, submission_id)
        if self.get_role(lecture_id).role == Scope.student and submission.user_id != self.user.id:
            raise HTTPError(HTTPStatus.NOT_FOUND)
        self.write_json(submission)

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Updates a specific submission and returns the updated entity.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        body = tornado.escape.json_decode(self.request.body)
        sub_model = SubmissionModel.from_dict(body)
        sub = self.get_submission(lecture_id, assignment_id, submission_id)
        # sub.date = sub_model.submitted_at
        # sub.assignid = assignment_id

        role = self.get_role(lecture_id)
        if role.role >= Scope.instructor:
            sub.user_id = sub_model.user_id
        sub.auto_status = sub_model.auto_status
        sub.manual_status = sub_model.manual_status
        sub.edited = sub_model.edited
        sub.feedback_status = sub_model.feedback_status
        if sub_model.score_scaling is not None and sub.score_scaling != sub_model.score_scaling:
            sub.score_scaling = sub_model.score_scaling
            if sub.grading_score is not None:
                # recalculate score based on new scaling factor
                sub.score = sub_model.score_scaling * sub.grading_score
        self.session.commit()
        self.write_json(sub)

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def delete(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Soft-deletes a specific submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: if submission has feedback, or the deadline has passed,
            or it has already been (soft-)deleted, or it belongs to another student,
            or it was not found in the given lecture and assignment.
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        self.validate_parameters()
        submission = self.get_submission(lecture_id, assignment_id, submission_id)

        if submission is None:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Submission to delete not found.")

        # Do not allow students to delete other users' submissions
        if self.get_role(lecture_id).role < Scope.instructor and submission.user_id != self.user.id:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Submission to delete not found.")

        if submission.feedback_status != FeedbackStatus.NOT_GENERATED:
            raise HTTPError(
                HTTPStatus.FORBIDDEN, reason="Only submissions without feedback can be deleted."
            )

        # if assignment has deadline and it has already passed
        if (
            submission.assignment.settings.deadline
            and submission.assignment.settings.deadline
            < datetime.datetime.now(datetime.timezone.utc)
        ):
            raise HTTPError(
                HTTPStatus.FORBIDDEN,
                reason="Submission can't be deleted, due date of assigment has passed.",
            )

        submission.deleted = DeleteState.deleted
        self.session.commit()


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/"
    + r"(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/logs\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionLogsHandler(GraderBaseHandler):
    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Returns logs of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission logs are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        logs = self.session.query(SubmissionLogs).get(submission_id)
        if logs is not None:
            self.write_json(logs.logs)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Logs of submission were not found")


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/"
    + r"(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/"
    + r"properties\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionPropertiesHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions/
    {submission_id}/properties.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Returns the properties of a submission,

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission or
        their properties are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        properties = self.session.get(SubmissionProperties, submission_id)
        if properties is not None and properties.properties is not None:
            # delete source cells from properties if user is student
            if self.get_role(lecture_id).role == Scope.student:
                model = GradeBookModel.from_dict(json.loads(properties.properties))
                for notebook in model.notebooks.values():
                    notebook.source_cells_dict = {}
                self.write(json.dumps(model.to_dict()))
            else:
                self.write(properties.properties)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Properties of submission were not found")

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Updates the properties of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id, submission_id)
        properties_string: str = self.request.body.decode("utf-8")

        try:
            gradebook = GradeBookModel.from_dict(json.loads(properties_string))
        except Exception as e:
            self.log.info(e)
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Cannot parse properties file!")

        score = gradebook.score
        submission.grading_score = score
        submission.score = submission.score_scaling * score

        properties = SubmissionProperties(properties=properties_string, sub_id=submission.id)

        self.session.merge(properties)

        if submission.feedback_status == FeedbackStatus.GENERATED:
            submission.feedback_status = FeedbackStatus.FEEDBACK_OUTDATED

        if submission.manual_status == ManualStatus.MANUALLY_GRADED:
            submission.manually_graded = ManualStatus.BEING_EDITED

        self.session.commit()
        self.write_json(submission)

    # TODO: not used, remove?
    def get_extra_credit(self, gradebook):
        extra_credit = 0
        for notebook in gradebook.notebooks.values():
            for grades in notebook.grades:
                extra_credit += grades.extra_credit if grades.extra_credit is not None else 0
        self.log.info("Extra credit is " + str(extra_credit))
        return extra_credit


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/"
    + r"(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/edit\/?",
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionEditHandler(GraderBaseHandler):
    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int, submission_id: int):
        """Creates or overwrites (resets) the repository which stores changes of
        submissions files
        :param lecture_id: lecture id
        :param assignment_id: assignment id
        :param submission_id: submission id
        :return:
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        self.validate_parameters()

        submission = self.get_submission(lecture_id, assignment_id, submission_id)
        if submission.commit_hash == INSTRUCTOR_SUBMISSION_COMMIT_CASH:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                reason="This repo cannot be edited or reset, because it was created by instructor",
            )

        assignment = submission.assignment
        lecture = assignment.lecture

        # Path to the (bare!) repository which will store edited submission files
        git_repo_path = self.construct_git_dir(
            repo_type=GitRepoType.EDIT,
            lecture=lecture,
            assignment=assignment,
            submission=submission,
        )

        # Path to repository of student which contains the submitted files
        submission_repo_path = os.path.join(
            self.gitbase, lecture.code, str(assignment.id), "user", submission.user.name
        )
        if not os.path.exists(submission_repo_path):
            raise HTTPError(
                HTTPStatus.BAD_REQUEST, reason="The user submission repository does not exist"
            )

        if os.path.exists(git_repo_path):
            shutil.rmtree(git_repo_path)

        # Creating bare repository
        if not os.path.exists(git_repo_path):
            os.makedirs(git_repo_path, exist_ok=True)

        await self._run_command_async(
            ["git", "init", "--bare", "--initial-branch=main"], git_repo_path
        )

        # Create temporary paths to copy the submission files in the edit repository
        tmp_path = os.path.join(
            self.application.grader_service_dir,
            "tmp",
            lecture.code,
            str(assignment.id),
            "edit",
            str(submission.id),
        )

        tmp_input_path = os.path.join(tmp_path, "input")

        tmp_output_path = os.path.join(tmp_path, "output")

        if os.path.exists(tmp_path):
            shutil.rmtree(tmp_path, ignore_errors=True)

        os.makedirs(tmp_input_path, exist_ok=True)

        # Init local repository
        await self._run_command_async(["git", "init", "--initial-branch=main"], tmp_input_path)

        # Pull user repository
        await self._run_command_async(
            ["git", "pull", str(submission_repo_path), "main"], tmp_input_path
        )
        self.log.info("Successfully cloned repo")

        # Checkout to correct submission commit
        await self._run_command_async(["git", "checkout", submission.commit_hash], tmp_input_path)
        self.log.info(f"Now at commit {submission.commit_hash}")

        # Copy files to output directory
        shutil.copytree(tmp_input_path, tmp_output_path, ignore=shutil.ignore_patterns(".git"))

        # Init local repository
        await self._run_command_async(["git", "init", "--initial-branch=main"], tmp_output_path)

        # Add edit remote
        await self._run_command_async(
            ["git", "remote", "add", "edit", str(git_repo_path)], tmp_output_path
        )
        self.log.info("Successfully added edit remote")

        # Switch to main
        await self._run_command_async(["git", "switch", "-c", "main"], tmp_output_path)
        self.log.info("Successfully switched to branch main")

        # Add files to staging
        await self._run_command_async(["git", "add", "-A"], tmp_output_path)
        self.log.info("Successfully added files to staging")

        # Commit Files
        await self._run_command_async(["git", "commit", "-m", "Initial commit"], tmp_output_path)
        self.log.info("Successfully commited files")

        # Push copied files
        await self._run_command_async(["git", "push", "edit", "main"], tmp_output_path)
        self.log.info("Successfully pushed copied files")

        submission.edited = True
        self.session.commit()
        self.write_json(submission)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/lti\/?",
    version_specifier=VersionSpecifier.ALL,
)
class LtiSyncHandler(GraderBaseHandler):
    cache_token = {"token": None, "ttl": datetime.datetime.now()}

    @authorize([Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int):
        """Starts the LTI sync process (if enabled).
        Request can have an optional parameter 'option'.
        if 'option' == 'latest': sync all latest submissions with feedback
           'option' == 'best': sync all best submissions with feedback
           'option' == 'selection': sync all submissions in body (default, if no param is set)

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters("option")
        lti_option = self.get_argument("option", "latest")

        assignment: Assignment = self.session.get(Assignment, assignment_id)
        if (
            (assignment is None)
            or (assignment.deleted == DeleteState.deleted)
            or (int(assignment.lectid) != int(lecture_id))
        ):
            self.log.error("Assignment with id " + str(assignment_id) + " was not found")
            return None
        lecture: Lecture = assignment.lecture

        # get all latest submissions with feedback
        if lti_option == "latest":
            submissions = self.get_latest_submissions(assignment_id, must_have_feedback=True)
        # get all best submissions with feedback
        elif lti_option == "best":
            submissions = self.get_best_submissions(assignment_id, must_have_feedback=True)
        else:
            # get submissions with given submission ids
            try:
                body = tornado.escape.json_decode(self.request.body)
                submission_ids: List[int] = body.get("submission_ids", [])

                if not submission_ids:
                    raise ValueError("No submission IDs provided")

                # Fetch and validate submissions
                submissions = (
                    self.session.query(Submission)
                    .filter(
                        Submission.id.in_(submission_ids),
                        Submission.auto_status == AutoStatus.AUTOMATICALLY_GRADED,
                        Submission.assignid == assignment_id,
                    )
                    .all()
                )
                if len(submissions) != len(submission_ids):
                    raise HTTPError(
                        HTTPStatus.BAD_REQUEST,
                        reason="Some submission IDs are invalid or do not belong to this assignment.",
                    )

            except Exception as e:
                err_msg = f"Could not process submission IDs: {e}"
                self.log.error(err_msg)
                raise HTTPError(HTTPStatus.BAD_REQUEST, reason=err_msg)

        lti_plugin = LTISyncGrades.instance()
        lecture_model = lecture.serialize()
        assignment_model = assignment.serialize()
        submissions_model = [sub.serialize_with_user() for sub in submissions]

        # check if the lti plugin is enabled
        if lti_plugin.check_if_lti_enabled(
            lecture_model, assignment_model, submissions_model, feedback_sync=False
        ):
            try:
                results = await lti_plugin.start(lecture_model, assignment_model, submissions_model)
                return self.write_json(results)
            except HTTPError as e:
                err_msg = f"Could not sync grades: {e.reason}"
                self.log.info(err_msg)
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=err_msg)
            except Exception as e:
                self.log.error("Could not sync grades: " + str(e))
                raise HTTPError(500, reason="An unexpected error occurred.")
        else:
            raise HTTPError(403, reason="LTI plugin is not enabled by administrator.")


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/count\/?"
)
class SubmissionCountHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/submissions/count.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Returns the count of submissions made by the student for an assignment, no matter if submission
         was deleted or not.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)

        role = self.get_role(lecture_id)

        usersubmissions_count = (
            self.session.query(Submission)
            .filter(Submission.assignid == assignment_id, Submission.user_id == role.user_id)
            .count()
        )

        self.write_json({"submission_count": usersubmissions_count})
        self.session.close()



import datetime
import json
import os
import time
from http import HTTPStatus
from urllib.parse import urlparse

import jwt
from tornado.escape import json_decode, url_escape
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPRequest
from tornado.web import HTTPError
from traitlets import Bool, Callable, Unicode, Union
from traitlets.config import SingletonConfigurable


def default_lti_username_match(member, submission, log) -> bool:
    return False


def default_enable_lti(lecture, assignment, submissions):
    return False


def default_enable_sync_on_feedback(lecture, assignment, submissions):
    return False


class LTISyncGrades(SingletonConfigurable):
    enabled = Union(
        [Bool(False), Callable(default_enable_lti)],
        allow_none=True,
        config=True,
        help="""
        Determines if the LTI Sync Grades plugin should be used, defaults to False.
        Is either a bool value or
        a function with the params (lecture, assignment, submissions) returning a bool.
        """,
    )

    sync_on_feedback = Union(
        [Bool(False), Callable(default_enable_sync_on_feedback)],
        allow_none=True,
        config=True,
        help="""
        Determines if submissions should be automatically synchronised, on feedback generation, defaults to False.
        Only synchronises scores when LTISyncGrades.enabled is True.
        Is either a bool value or
        a function with the params (lecture, assignment, submissions) returning a bool.
        """,
    )
    client_id = Unicode(None, config=True, allow_none=True)
    token_url = Unicode(None, config=True, allow_none=True)
    username_match = Callable(
        default_value=default_lti_username_match,
        config=True,
        allow_none=True,
        help="Function used to match lti member object with submission object to. Is given member, submission object and log as params and returns boolean if it is the users submission",
    )

    token_private_key = Union(
        [Unicode(os.environ.get("LTI_PRIVATE_KEY", None)), Callable(None)],
        allow_none=True,
        config=True,
        help="""
        Private Key used to encrypt bearer token request
        """,
    )
    resolve_lti_urls = Callable(
        default_value=None,
        config=True,
        allow_none=True,
        help="Returns membership and lineitem URL needed for grade sync",
    )

    # cache for lti token
    cache_token = {"token": None, "ttl": datetime.datetime.now()}

    def check_if_lti_enabled(self, lecture, assignment, submissions, feedback_sync):
        if callable(self.enabled):
            enable_lti = self.enabled(lecture, assignment, submissions)
        else:
            enable_lti = self.enabled

        if enable_lti:
            if feedback_sync:
                if self.sync_on_feedback:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    async def start(self, lecture, assignment, submissions):
        self.log.info("LTI: start grade sync")
        if len(submissions) == 0:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="No submissions to sync")

        # 1. request bearer token
        self.log.debug("LTI: request bearer token")
        stamp = datetime.datetime.now()
        if self.cache_token["token"] and self.cache_token["ttl"] > stamp - datetime.timedelta(
            minutes=50
        ):
            token = self.cache_token["token"]
        else:
            token = await self.request_bearer_token()
            self.cache_token["token"] = token
            self.cache_token["ttl"] = datetime.datetime.now()

        # 2. resolve lti urls
        self.log.debug("LTI: resolve lti url")
        try:
            lti_urls = self.resolve_lti_urls(lecture, assignment, submissions)
            self.log.debug(f"LTI membership and lineitems URL: {lti_urls}")
            lineitems_url = lti_urls["lineitems_url"]
            membership_url = lti_urls["membership_url"]
        except Exception as e:
            self.log.error(e)
            raise e
        # 3. get all members
        self.log.debug("LTI: request all members of lti course")
        httpclient = AsyncHTTPClient()
        try:
            response = await httpclient.fetch(
                HTTPRequest(
                    url=membership_url,
                    method="GET",
                    headers={
                        "Authorization": "Bearer " + self.cache_token["token"],
                        "Accept": "application/vnd.ims.lti-nrps.v2.membershipcontainer+json",
                    },
                )
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to get users of course:" + e.response.reason)
        members = json_decode(response.body)["members"]

        # 4. match usernames of submissions to lti memberships
        # and generate for each submission a request body -> grades list
        self.log.debug("LTI: match grader usernames with lti identifier")
        grades = []
        syncable_user_count = 0
        for submission in submissions:
            for member in members:
                if self.username_match(member, submission, self.log):
                    syncable_user_count += 1
                    grades.append(
                        self.build_grade_publish_body(
                            member["user_id"], submission["score"], float(assignment["points"])
                        )
                    )
        self.log.info(f"LTI: matched {syncable_user_count} users")
        # 6. get all lineitems
        self.log.debug("LTI: resolve lti url")
        try:
            response = await httpclient.fetch(
                HTTPRequest(
                    url=lineitems_url,
                    method="GET",
                    headers={
                        "Authorization": "Bearer " + self.cache_token["token"],
                        "Accept": "application/vnd.ims.lis.v2.lineitemcontainer+json",
                    },
                )
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to get lineitems of course:" + e.response.reason)
        lineitems = json_decode(response.body)
        self.log.debug(f"LTI found lineitems: {lineitems}")

        # 7. check if a lineitem with assignment name exists
        lineitem = None
        for item in lineitems:
            if item["label"] == assignment["name"]:
                # lineitem found
                self.log.debug(f"LTI found lineitem: {item}")
                lineitem = item
                break

        # 8. if does not exist, create a lineitem with the assignment name
        if lineitem is None:
            lineitem_body = {
                "scoreMaximum": float(assignment["points"]),
                "label": assignment["name"],
                "resourceId": assignment["id"],
                "tag": "grade",
                "startDateTime": str(datetime.datetime.now()),
                "endDateTime": str(datetime.date.today() + datetime.timedelta(days=1, hours=1)),
            }
            try:
                response = await httpclient.fetch(
                    HTTPRequest(
                        url=lineitems_url,
                        method="POST",
                        body=json.dumps(lineitem_body),
                        headers={
                            "Authorization": "Bearer " + self.cache_token["token"],
                            "Content-Type": "application/vnd.ims.lis.v2.lineitem+json",
                        },
                    )
                )
            except HTTPClientError as e:
                self.log.error(e.response)
                raise HTTPError(
                    e.code, reason="Unable to create new lineitem in course:" + e.response.reason
                )
            # due to different "interpretations" of the ims lti standard,
            # the response is sometimes a list containing the lineitem or
            # just the lineitem json
            try:
                lineitem_response = json_decode(response.body)
            except Exception as e:
                self.log.error("LTI: could not decode lineitem request response")
                raise e
            if isinstance(lineitem_response, list):
                lineitem = lineitem_response[0]
            elif isinstance(lineitem_response, dict):
                lineitem = lineitem_response
            else:
                self.log.error("LTI: lineitem request response does not match dict or list")
                raise HTTPError(
                    HTTPStatus.UNPROCESSABLE_ENTITY,
                    "lineitem request response does not match dict or list",
                )

        # 9. push grades to lineitem
        url_parsed = urlparse(lineitem["id"])
        lineitem = url_parsed._replace(path=url_parsed.path + "/scores").geturl()
        self.log.debug("LTI: start sending grades to LTI course")
        synced_user = 0
        for grade in grades:
            try:
                response = await httpclient.fetch(
                    HTTPRequest(
                        url=lineitem,
                        method="POST",
                        body=json.dumps(grade),
                        headers={
                            "Authorization": "Bearer " + self.cache_token["token"],
                            "Content-Type": "application/vnd.ims.lis.v1.score+json",
                        },
                    )
                )
                synced_user += 1
            except HTTPClientError as e:
                self.log.error(e.response)
        self.log.info("LTI Grade Sync finished successfully")
        return {"syncable_users": syncable_user_count, "synced_user": synced_user}

    def build_grade_publish_body(self, uid: str, score: float, max_score: float):
        return {
            "timestamp": str(
                datetime.datetime.now().astimezone().replace(microsecond=0).isoformat()
            ),
            "scoreGiven": score,
            "comment": "Automatically synced",
            "scoreMaximum": max_score,
            "activityProgress": "Submitted",
            "gradingProgress": "FullyGraded",
            "userId": uid,
        }

    async def request_bearer_token(self):
        # get config variables
        if self.client_id is None:
            raise HTTPError(
                HTTPStatus.NOT_FOUND,
                reason="Unable to request bearer token: client_id is not set in grader config",
            )
        if self.token_url is None:
            raise HTTPError(
                HTTPStatus.NOT_FOUND,
                reason="Unable to request bearer token: token_url is not set in grader config",
            )

        private_key = self.token_private_key
        if private_key is None:
            raise HTTPError(
                HTTPStatus.NOT_FOUND,
                reason="Unable to request bearer token: token_private_key is not set in grader config",
            )
        if callable(private_key):
            private_key = private_key()
        headers = {"typ": "JWT", "alg": "RS256"}
        payload = {
            "iss": "grader-service",
            "sub": self.client_id,
            "aud": [self.token_url],
            "iat": str(int(time.time())),
            "exp": str(int(time.time()) + 60),
            "jti": str(int(time.time())) + "123",
        }
        try:
            encoded = jwt.encode(payload, private_key, algorithm="RS256", headers=headers)
        except Exception as e:
            raise HTTPError(
                HTTPStatus.UNPROCESSABLE_ENTITY, reason=f"Unable to encode payload: {str(e)}"
            )
        scopes = [
            "https://purl.imsglobal.org/spec/lti-ags/scope/score",
            "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem",
            "https://purl.imsglobal.org/spec/lti-nrps/scope/contextmembership.readonly",
        ]
        scopes = url_escape(" ".join(scopes))
        data = (
            f"grant_type=client_credentials&client_assertion_type=urn%3Aietf%3Aparams%3Aoauth%3Aclient-assertion"
            f"-type%3Ajwt-bearer&client_assertion={encoded}&scope={scopes}"
        )
        httpclient = AsyncHTTPClient()
        try:
            response = await httpclient.fetch(
                HTTPRequest(
                    url=self.token_url,
                    method="POST",
                    body=data,
                    headers={
                        "Content-Type": "application/x-www-form-urlencoded",
                        "Content-Length": len(data),
                    },
                )
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason="Unable to request token:" + e.response.reason)
        return json_decode(response.body)["access_token"]



import os
import subprocess
from typing import Any, List, Optional

from traitlets import Unicode, validate
from traitlets.config import LoggingConfigurable

from grader_service.autograding.utils import executable_validator
from grader_service.handlers.handler_utils import GitRepoType
from grader_service.orm import Assignment, Lecture, Submission


class GitSubmissionManager(LoggingConfigurable):
    """
    Handles git-related operations performed by autograder executors:
    pulling from an input repo, and committing and pushing to the output repo.
    """

    git_executable = Unicode("git", allow_none=False).tag(config=True)
    input_repo_type = GitRepoType.USER
    output_repo_type = GitRepoType.AUTOGRADE

    def __init__(self, grader_service_dir: str, submission: Submission, **kwargs: Any):
        super().__init__(**kwargs)
        self.grader_service_dir = grader_service_dir
        self.submission = submission

        if self.input_repo_type == GitRepoType.USER and self.submission.edited:
            # User's submission was edited by the instructor - repo type has to be adjusted
            self.input_repo_type = GitRepoType.EDIT

        self.input_branch = "main"
        self.output_branch = f"submission_{self.submission.commit_hash}"

    def _get_repo_path(self, repo_type: GitRepoType) -> str:
        """Determines the Git repository path for the submission."""
        assignment: Assignment = self.submission.assignment
        lecture: Lecture = assignment.lecture
        repo_name = self.submission.user.name

        base_repo_path = os.path.join(
            self.grader_service_dir, "git", lecture.code, str(assignment.id), repo_type
        )
        if repo_type in [GitRepoType.AUTOGRADE, GitRepoType.FEEDBACK]:
            path = os.path.join(base_repo_path, "user", repo_name)
        elif repo_type == GitRepoType.EDIT:
            path = os.path.join(base_repo_path, str(self.submission.id))
        elif repo_type == GitRepoType.USER:
            path = os.path.join(base_repo_path, repo_name)
        else:
            raise ValueError(f"Cannot determine repo path for repo type {repo_type}")

        path = os.path.normpath(path)

        if not path.startswith(self.grader_service_dir):
            self.log.error(
                f"Invalid repo path: {path}. Possibly suspicious values: "
                f"lecture code: '{lecture.code}' or user name: '{repo_name}'"
            )
            raise PermissionError("Invalid repository path.")

        return path

    def pull_submission(self, input_path: str) -> None:
        """Inits and pulls the submission repository into the input path.

        :param input_path: The directory where the input repo will be created.
        """
        input_repo_path = self._get_repo_path(self.input_repo_type)

        self.log.info(f"Pulling repo {input_repo_path} into input directory")
        commands = [
            [self.git_executable, "init"],
            [self.git_executable, "pull", input_repo_path, self.input_branch],
        ]

        # When autograding a user's submission, check out to the commit of submission
        if self.input_repo_type == GitRepoType.USER:
            commands.append([self.git_executable, "checkout", self.submission.commit_hash])

        for cmd in commands:
            self._run_git(cmd, input_path)

        self.log.info("Successfully cloned repo.")

    def _set_up_output_repo(self, output_path: str) -> None:
        """Initializes the output repo and switches to a separate branch named
        after the commit hash of the submission."""
        output_repo_path = self._get_repo_path(self.output_repo_type)

        if not os.path.exists(output_repo_path):
            os.makedirs(output_repo_path)
            self._run_git([self.git_executable, "init", "--bare", output_repo_path], output_path)

        self.log.info(f"Initialising repo at {output_path}")
        self._run_git([self.git_executable, "init"], output_path)
        self.log.info(f"Creating the new branch {self.output_branch} and switching to it")
        command = [self.git_executable, "switch", "-c", self.output_branch]
        self._run_git(command, output_path)
        self.log.info(f"Now at branch {self.output_branch}")

    def _commit_files(self, filenames: List[str], output_path: str) -> None:
        """
        Commits the provided files in the repo at `output_path`.
        """
        self.log.info(f"Committing files in {output_path}")

        if not filenames:
            self.log.info("No files to commit.")
            return

        # Make sure we do not commit the gradebook.json
        filenames = [f for f in filenames if f != "gradebook.json"]

        self._run_git([self.git_executable, "add", "--", *filenames], output_path)
        self._run_git(
            [self.git_executable, "commit", "-m", self.submission.commit_hash], output_path
        )

    def push_results(self, filenames: List[str], output_path: str) -> None:
        """Creates the output repository, commits and pushes the changes."""
        self._set_up_output_repo(output_path)
        self._commit_files(filenames, output_path)

        output_repo_path = self._get_repo_path(self.output_repo_type)

        self.log.info(f"Pushing to {output_repo_path} at branch {self.output_branch}")
        self._run_git(
            [self.git_executable, "push", "-uf", output_repo_path, self.output_branch], output_path
        )
        self.log.info("Pushing complete")

    def _run_git(self, command: list[str], cwd: Optional[str]) -> None:
        """
        Execute a git command as a subprocess.

        Args:
            command: The git command to execute, as a list of strings.
            cwd: The working directory the subprocess should run in.
        Raises:
            `subprocess.CalledProcessError`: if `subprocess.run` fails.
            Any other exception thrown while running the subprocess is logged and also re-raised.
        """
        assert command[0] == self.git_executable, f"Not a git command: {command}"
        self.log.debug('Running "%s"', " ".join(command))
        try:
            subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=cwd,
                text=True,  # Decodes output to string
                check=True,  # Raises a CalledProcessError on non-zero exit code
            )
        except subprocess.CalledProcessError as e:
            self.log.error(e.stderr)
            raise
        except Exception as e:
            self.log.error(e)
            raise

    @validate("git_executable")
    def _validate_executable(self, proposal):
        return executable_validator(proposal)



# Copyright (c) 2025, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import io
import logging
import os
import shutil
import stat
import sys
from contextlib import contextmanager
from typing import Iterator

from traitlets import TraitError
from traitlets.config.configurable import LoggerType


def rmtree(path: str) -> None:
    """Python version-independent function to remove a directory tree.

    Exceptions are handled by the `rm_error` handler.
    In `shutil.rmtree`, `onerror` is deprecated in 3.12 and `onexc` is introduced instead.

    :param path: Path pointing to the directory to remove.
    :type path: str
    """
    if sys.version_info >= (3, 12):
        shutil.rmtree(path, onexc=rm_error)
    else:
        shutil.rmtree(path, onerror=rm_error)


def rm_error(func, path, exc_info):
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


@contextmanager
def collect_logs(logger: LoggerType) -> Iterator[io.StringIO]:
    """
    A context manager for collecting logs from a logger.

    Adds a `StreamHandler` to the `logger` and yields a log stream, whose content
    can then be retrieved with `log_stream.getvalue().

    Example usage for collecting `Autograde` logs:

    .. code-block:: python

        autograder = Autograde(...)
        with collect_logs(autograder.log) as log_stream:
            autograder.start()
            grading_logs = log_stream.getvalue()

    """
    log_stream = io.StringIO()
    handler = logging.StreamHandler(log_stream)
    handler.setFormatter(
        logging.Formatter(
            fmt="[%(asctime)s] [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
    )
    logger.addHandler(handler)
    try:
        yield log_stream
    finally:
        logger.removeHandler(handler)
        handler.close()


def executable_validator(proposal: dict) -> str:
    exec: str = proposal["value"]
    if shutil.which(exec) is None:
        raise TraitError(f"The executable is not valid: {exec}")
    return exec



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class ManualStatus(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    """
    allowed enum values
    """
    NOT_GRADED = 'not_graded'
    MANUALLY_GRADED = 'manually_graded'
    BEING_EDITED = 'being_edited'
    def __init__(self):  # noqa: E501
        """ManualStatus - a model defined in OpenAPI

        """
        self.openapi_types = {
        }

        self.attribute_map = {
        }

    @classmethod
    def from_dict(cls, dikt) -> 'ManualStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ManualStatus of this ManualStatus.  # noqa: E501
        :rtype: ManualStatus
        """
        return util.deserialize_model(dikt, cls)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from grader_service.orm.api_token import APIToken
from grader_service.orm.assignment import Assignment
from grader_service.orm.base import Base
from grader_service.orm.lecture import Lecture
from grader_service.orm.oauthclient import OAuthClient
from grader_service.orm.oauthcode import OAuthCode
from grader_service.orm.submission import Submission
from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties
from grader_service.orm.takepart import Role
from grader_service.orm.user import User

__all__ = [
    "Lecture",
    "User",
    "Role",
    "Submission",
    "Assignment",
    "Base",
    "OAuthCode",
    "OAuthClient",
    "APIToken",
    "SubmissionLogs",
    "SubmissionProperties",
]



import sys

from grader_service._version import __version__
from grader_service.main import GraderService

__all__ = ["GraderService", "__version__"]

main = GraderService.launch_instance

if __name__ == "__main__":
    main(sys.argv)



"""Base Authenticator class"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
import re

from traitlets import Any, Bool, Callable, Dict, Integer, Set, Unicode, Union, default, observe
from traitlets.config import LoggingConfigurable

from grader_service.utils import maybe_future, url_path_join

from .login import LoginHandler


class Authenticator(LoggingConfigurable):
    """Base class for implementing an authentication provider for Grader Service."""

    login_redirect_url = Union(
        [Unicode(), Callable()],
        default_value="/",
        allow_none=False,
        help="""
        The default URL to redirect users when they successfully logged in.

        By default, this will likely be the URL of the JupyterHub server.

        Can be a Unicode string (e.g. '/hub/home') or a callable based on the handler object:

        ::

            def login_redirect_url_fn(handler):
                user = handler.current_user
                if user and user.admin:
                    return '/hub/admin'
                return '/hub/home'

            c.Authenticator.login_redirect_url = login_redirect_url_fn
        """,
    ).tag(config=True)

    enable_auth_state = Bool(
        False,
        config=True,
        help="""Enable persisting auth_state (if available).

        auth_state will be encrypted and stored in the Hub's database.
        This can include things like authentication tokens, etc.
        to be passed to Spawners as environment variables.

        Encrypting auth_state requires the cryptography package.

        Additionally, the JUPYTERHUB_CRYPT_KEY environment variable must
        contain one (or more, separated by ;) 32B encryption keys.
        These can be either base64 or hex-encoded.

        If encryption is unavailable, auth_state cannot be persisted.

        """,
    )

    auth_refresh_age = Integer(
        300,
        config=True,
        help="""The max age (in seconds) of authentication info
        before forcing a refresh of user auth info.

        Refreshing auth info allows, e.g. requesting/re-validating auth tokens.

        See :meth:`.refresh_user` for what happens when user auth info is refreshed
        (nothing by default).
        """,
    )

    refresh_pre_spawn = Bool(
        False,
        config=True,
        help="""Force refresh of auth prior to spawn.

        This forces :meth:`.refresh_user` to be called prior to launching
        a server, to ensure that auth state is up-to-date.

        This can be important when e.g. auth tokens that may have expired
        are passed to the spawner via environment variables from auth_state.

        If refresh_user cannot refresh the user auth data,
        launch will fail until the user logs in again.
        """,
    )

    admin_users = Set(
        help="""
        .. warning::
            Admin users do currently not have any special privileges
            in the Grader Service, but this may change in the future.

        Set of users that will have admin rights on the Grader Service instance.
        Admin access should be treated the same way root access is.

        Defaults to an empty set, in which case no user has admin access.
        """
    ).tag(config=True)

    any_allow_config = Bool(
        False,
        help="""Is there any allow config?

        Deployments can set this explicitly to True to suppress
        the "No allow config found" warning.

        Will be True if any config tagged with `.tag(allow_config=True)`
        or starts with `allow` is truthy.

        """,
    ).tag(config=True)

    @default("any_allow_config")
    def _default_any_allowed(self):
        for trait_name, trait in self.traits(config=True).items():
            if trait.metadata.get("allow_config", False) or trait_name.startswith("allow"):
                # this is only used for a helpful warning, so not the biggest deal if it's imperfect
                if getattr(self, trait_name):
                    return True
        return False

    def check_allow_config(self):
        """Log a warning if no allow config can be found.

        Could get a false positive if _only_ unrecognized allow config is used.
        Authenticators can apply `.tag(allow_config=True)` to label this config
        to make sure it is found.

        Subclasses can override to perform additonal checks and warn about likely
        authenticator configuration problems.

        .. versionadded:: 5.0
        """
        if not self.any_allow_config:
            self.log.warning(
                "No allow config found, it's possible that nobody can login to your Hub!\n"
                "You can set `c.Authenticator.allow_all = True` to allow any user who can login to access the Hub,\n"
                "or e.g. `allowed_users` to a set of users who should have access.\n"
                "You may suppress this warning by setting c.Authenticator.any_allow_config = True."
            )

    allowed_users = Set(
        help="""
        Set of usernames that are allowed to log in.

        Use this to limit which authenticated users may login.
        Default behavior: only users in this set are allowed.

        If empty, does not perform any restriction,
        in which case any authenticated user is allowed.

        Authenticators may extend :meth:`.Authenticator.check_allowed`
        to combine `allowed_users` with other configuration
        to either expand or restrict access.

        """
    ).tag(config=True)

    allow_all = Bool(
        False,
        config=True,
        help="""
        Allow every user who can successfully authenticate to access JupyterHub.

        False by default, which means for most Authenticators,
        _some_ allow-related configuration is required to allow users to log in.

        Authenticator subclasses may override the default with e.g.::

            @default("allow_all")
            def _default_allow_all(self):
                # if _any_ auth config (depends on the Authenticator)
                if self.allowed_users or self.allowed_groups or self.allow_existing_users:
                    return False
                else:
                    return True
        """,
    ).tag(allow_config=True)

    allow_existing_users = Bool(
        # dynamic default computed from allowed_users
        config=True,
        help="""
        Allow existing users to login.

        Defaults to True if `allowed_users` is set for historical reasons, and
        False otherwise.

        With this enabled, all users present in the Grader Service database are allowed to login.
        This has the effect of any user who has _previously_ been allowed to login
        via any means will continue to be allowed until the user is deleted.
        """,
    ).tag(allow_config=True)

    @default("allow_existing_users")
    def _allow_existing_users_default(self):
        """
        Computes the default value of allow_existing_users based on if
        allowed_users to align with original behavior not introduce a breaking
        change.
        """
        if self.allowed_users:
            return True
        return False

    blocked_users = Set(
        help="""
        Set of usernames that are not allowed to log in.

        Use this with supported authenticators to restrict which users can not log in. This is an
        additional block list that further restricts users, beyond whatever restrictions the
        authenticator has in place.

        If empty, does not perform any additional restriction.
        """
    ).tag(config=True)

    otp_prompt = Any(
        "OTP:",
        help="""
        The prompt string for the extra OTP (One Time Password) field.
        """,
    ).tag(config=True)

    request_otp = Bool(
        False,
        config=True,
        help="""
        Prompt for OTP (One Time Password) in the login form.
        """,
    )

    @observe("allowed_users")
    def _check_allowed_users(self, change):
        short_names = [name for name in change["new"] if len(name) <= 1]
        if short_names:
            sorted_names = sorted(short_names)
            single = "".join(sorted_names)
            string_set_typo = "set('%s')" % single
            self.log.warning(
                "Allowed set contains single-character names: %s; did you mean set([%r]) instead of %s?",
                sorted_names[:8],
                single,
                string_set_typo,
            )

    custom_html = Unicode(
        help="""
        HTML form to be overridden by authenticators if they want a custom authentication form.

        Defaults to an empty string, which shows the default username/password form.
        """
    )

    def get_custom_html(self, base_url):
        """Get custom HTML for the authenticator."""
        return self.custom_html

    login_service = Unicode(
        help="""
        Name of the login service that this authenticator is providing using to authenticate users.

        Example: GitHub, MediaWiki, Google, etc.

        Setting this value replaces the login form with a "Login with <login_service>" button.

        Any authenticator that redirects to an external service (e.g. using OAuth) should set this.
        """
    )

    username_pattern = Unicode(
        help="""
        Regular expression pattern that all valid usernames must match.

        If a username does not match the pattern specified here, authentication will not be attempted.

        If not set, allow any username.
        """
    ).tag(config=True)

    @observe("username_pattern")
    def _username_pattern_changed(self, change):
        if not change["new"]:
            self.username_regex = None
        self.username_regex = re.compile(change["new"])

    username_regex = Any(
        help="""
        Compiled regex kept in sync with `username_pattern`
        """
    )

    def validate_username(self, username):
        """Validate a normalized username

        Return True if username is valid, False otherwise.
        """
        if "/" in username:
            # / is not allowed in usernames
            return False
        if not username:
            # empty usernames are not allowed
            return False
        if username != username.strip():
            # starting/ending with space is not allowed
            return False
        if not self.username_regex:
            return True
        return bool(self.username_regex.match(username))

    username_map = Dict(
        help="""Dictionary mapping authenticator usernames to JupyterHub users.

        Primarily used to normalize OAuth user names to local users.
        """
    ).tag(config=True)

    delete_invalid_users = Bool(
        False,
        config=True,
        help="""Delete any users from the database that do not pass validation

        When JupyterHub starts, `.add_user` will be called
        on each user in the database to verify that all users are still valid.

        If `delete_invalid_users` is True,
        any users that do not pass validation will be deleted from the database.
        Use this if users might be deleted from an external system,
        such as local user accounts.

        If False (default), invalid users remain in the Hub's database
        and a warning will be issued.
        This is the default to avoid data loss due to config changes.
        """,
    )

    post_auth_hook = Any(
        config=True,
        help="""
        A hook function that you can implement to do some
        bootstrapping work during authentication. For example, loading user account
        details from an external system.

        This function is called after the user has passed all authentication checks
        and is ready to successfully authenticate. This function must return the
        auth_model dict reguardless of changes to it.
        The hook is called with 3 positional arguments: `(authenticator, handler, auth_model)`.

        This may be a coroutine.

        Example::

            import os
            import pwd
            def my_hook(authenticator, handler, auth_model):
                user_data = pwd.getpwnam(auth_model['name'])
                spawn_data = {
                    'pw_data': user_data
                    'gid_list': os.getgrouplist(auth_model['name'], user_data.pw_gid)
                }

                if auth_model['auth_state'] is None:
                    auth_model['auth_state'] = {}
                auth_model['auth_state']['spawn_data'] = spawn_data

                return auth_model

            c.Authenticator.post_auth_hook = my_hook

        """,
    )

    async def run_post_auth_hook(self, handler, auth_model):
        """
        Run the post_auth_hook if defined

        Args:
            handler (tornado.web.RequestHandler): the current request handler
            auth_model (dict): User authentication data dictionary. Contains the
                username ('name'), admin status ('admin'), and auth state dictionary ('auth_state').
        Returns:
            auth_model (dict):
                The hook must always return the auth_model dict
        """
        if self.post_auth_hook is not None:
            auth_model = await maybe_future(self.post_auth_hook(self, handler, auth_model))
        return auth_model

    def normalize_username(self, username):
        """Normalize the given username and return it

        Override in subclasses if usernames need different normalization rules.

        The default attempts to lowercase the username and apply `username_map` if it is
        set.
        """
        username = username.lower()
        username = self.username_map.get(username, username)
        return username

    def check_allowed(self, username, authentication=None):
        """Check if a username is allowed to authenticate based on configuration

        Return True if username is allowed, False otherwise.

        No allowed_users set means any username is allowed.

        Names are normalized *before* being checked against the allowed set.

        Args:
            username (str):
                The normalized username
            authentication (dict):
                The authentication model, as returned by `.authenticate()`.
        Returns:
            allowed (bool):
                Whether the user is allowed
        Raises:
            web.HTTPError(403):
                Raising HTTPErrors directly allows customizing the message shown to the user.
        """
        if self.allow_all:
            return True
        return username in self.allowed_users

    def check_blocked_users(self, username, authentication=None):
        """Check if a username is blocked to authenticate based on Authenticator.blocked_users configuration

        Return True if username is allowed, False otherwise.
        No block list means any username is allowed.

        Names are normalized *before* being checked against the block list.

        Args:
            username (str):
                The normalized username
            authentication (dict):
                The authentication model, as returned by `.authenticate()`.
        Returns:
            allowed (bool):
                Whether the user is allowed
        Raises:
            web.HTTPError(403, message):
                Raising HTTPErrors directly allows customizing the message shown to the user.
        """
        if not self.blocked_users:
            # No block list means any name is allowed
            return True
        return username not in self.blocked_users

    async def get_authenticated_user(self, handler, data):
        """Authenticate the user who is attempting to log in

        Returns user dict if successful, None otherwise.

        This calls `authenticate`, which should be overridden in subclasses,
        normalizes the username if any normalization should be done,
        and then validates the name in the allowed set.

        This is the outer API for authenticating a user.
        Subclasses should not override this method.

        The various stages can be overridden separately:
         - `authenticate` turns formdata into a username
         - `normalize_username` normalizes the username
         - `check_blocked_users` check against the blocked usernames
         - `allow_all` is checked
         - `check_allowed` checks against the allowed usernames
         - `is_admin` check if a user is an admin

        """
        authenticated = await maybe_future(self.authenticate(handler, data))
        if authenticated is None:
            return
        if isinstance(authenticated, dict):
            if "name" not in authenticated:
                raise ValueError("user missing a name: %r" % authenticated)
        else:
            authenticated = {"name": authenticated}
        authenticated.setdefault("auth_state", None)
        # Leave the default as None, but reevaluate later post-allowed-check
        authenticated.setdefault("admin", None)

        # normalize the username
        authenticated["name"] = username = self.normalize_username(authenticated["name"])
        if not self.validate_username(username):
            self.log.warning("Disallowing invalid username %r.", username)
            return

        blocked_pass = await maybe_future(self.check_blocked_users(username, authenticated))

        if not blocked_pass:
            self.log.warning("User %r blocked. Stop authentication", username)
            return

        allowed_pass = self.allow_all
        if not allowed_pass:
            allowed_pass = await maybe_future(self.check_allowed(username, authenticated))

        if allowed_pass:
            if authenticated["admin"] is None:
                authenticated["admin"] = await maybe_future(self.is_admin(handler, authenticated))

            authenticated = await self.run_post_auth_hook(handler, authenticated)

            return authenticated
        else:
            self.log.warning("User %r not allowed.", username)
            return

    async def refresh_user(self, user, handler=None):
        """Refresh auth data for a given user

        Allows refreshing or invalidating auth data.

        Only override if your authenticator needs
        to refresh its data about users once in a while.

        Args:
            user (User): the user to refresh
            handler (tornado.web.RequestHandler or None): the current request handler
        Returns:
            auth_data (bool or dict):
                Return **True** if auth data for the user is up-to-date
                and no updates are required.

                Return **False** if the user's auth data has expired,
                and they should be required to login again.

                Return a **dict** of auth data if some values should be updated.
                This dict should have the same structure as that returned
                by :meth:`.authenticate()` when it returns a dict.
                Any fields present will refresh the value for the user.
                Any fields not present will be left unchanged.
                This can include updating `.admin` or `.auth_state` fields.
        """
        return True

    def is_admin(self, handler, authentication):
        """Authentication helper to determine a user's admin status.

        Args:
            handler (tornado.web.RequestHandler): the current request handler
            authentication: The authentication dict generated by `authenticate`.
        Returns:
            admin_status (Bool or None):
                The admin status of the user, or None if it could not be
                determined or should not change.
        """
        return True if authentication["name"] in self.admin_users else None

    async def authenticate(self, handler, data):
        """Authenticate a user with login form data

        This must be a coroutine.

        It must return the username on successful authentication,
        and return None on failed authentication.

        Subclasses can also raise a `web.HTTPError(403, message)`
        in order to halt the authentication process
        and customize the error message that will be shown to the user.
        This error may be raised anywhere in the authentication process
        (`authenticate`, `check_allowed`, `check_blocked_users`).

        Checking allowed_users/blocked_users is handled separately by the caller.

        Args:
            handler (tornado.web.RequestHandler): the current request handler
            data (dict): The formdata of the login form.
                         The default form has 'username' and 'password' fields.
        Returns:
            user (str or dict or None):
                The username of the authenticated user,
                or None if Authentication failed.

                The Authenticator may return a dict instead, which MUST have a
                key `name` holding the username, and MAY have additional keys:

                - `auth_state`, a dictionary of auth state that will be persisted;
                - `admin`, the admin setting value for the user
                - `groups`, the list of group names the user should be a member of,
                  if Authenticator.manage_groups is True.
                  `groups` MUST always be present if manage_groups is enabled.
        Raises:
            web.HTTPError(403):
                Raising errors directly allows customizing the message shown to the user.
        """

    async def load_managed_roles(self):
        """Load roles managed by authenticator.

        Returns a list of predefined role dictionaries to load at startup,
        following the same format as `JupyterHub.load_roles`.

        """
        if not self.manage_roles:
            raise ValueError("Managed roles can only be loaded when `manage_roles` is True")
        if self.reset_managed_roles_on_startup:
            raise NotImplementedError(
                "When `reset_managed_roles_on_startup` is used, the `load_managed_roles()`"
                " method must have a non-default implementation, because using the default"
                " implementation would remove all managed roles and role assignments."
            )
        return []

    def pre_spawn_start(self, user, spawner):
        """Hook called before spawning a user's server

        Can be used to do auth-related startup, e.g. opening PAM sessions.
        """

    def post_spawn_stop(self, user, spawner):
        """Hook called after stopping a user container

        Can be used to do auth-related cleanup, e.g. closing PAM sessions.
        """

    def add_user(self, user):
        """Hook called when a user is added to JupyterHub

        This is called:
         - When a user first authenticates, _after_ all allow and block checks have passed
         - When the hub restarts, for all users in the database (i.e. users previously allowed)
         - When a user is added to the database, either via configuration or REST API

        This method may be a coroutine.

        By default, this adds the user to the allowed_users set if
        allow_existing_users is true.

        Subclasses may do more extensive things, such as creating actual system users,
        but they should call super to ensure the allowed_users set is updated.

        Note that this should be idempotent, since it is called whenever the hub restarts
        for all users.

        Args:
            user (User): The User wrapper object
        """
        if not self.validate_username(user.name):
            raise ValueError("Invalid username: %s" % user.name)
        if self.allow_existing_users and not self.allow_all:
            self.allowed_users.add(user.name)

    def delete_user(self, user):
        """Hook called when a user is deleted

        Removes the user from the allowed_users set.
        Subclasses should call super to ensure the allowed_users set is updated.

        Args:
            user (User): The User wrapper object
        """
        self.allowed_users.discard(user.name)

    # TODO: remove manage_groups since we only have roles
    manage_groups = Bool(
        False,
        config=True,
        help="""Let authenticator manage user groups

        If True, Authenticator.authenticate and/or .refresh_user
        may return a list of group names in the 'groups' field,
        which will be assigned to the user.

        All group-assignment APIs are disabled if this is True.
        """,
    )
    manage_roles = Bool(
        False,
        config=True,
        help="""Let authenticator manage roles

        If True, Authenticator.authenticate and/or .refresh_user
        may return a list of roles in the 'roles' field,
        which will be added to the database.

        When enabled, all role management will be handled by the
        authenticator.

        """,
    )
    reset_managed_roles_on_startup = Bool(
        False,
        config=True,
        help="""Reset managed roles to result of `load_managed_roles()` on startup.

        If True:
          - stale managed roles will be removed,
          - stale assignments to managed roles will be removed.

        Any role not present in `load_managed_roles()` will be considered 'stale'.

        The 'stale' status for role assignments is also determined from `load_managed_roles()` result:

        - user role assignments status will depend on whether the `users` key is defined or not:

          * if a list is defined under the `users` key and the user is not listed, then the user
          role assignment will be considered 'stale',
          * if the `users` key is not provided, the user role assignment will be preserved;
        - service and group role assignments will be considered 'stale':

          * if not included in the `services` and `groups` list,
          * if the `services` and `groups` keys are not provided.

        """,
    )
    auto_login = Bool(
        False,
        config=True,
        help="""Automatically begin the login process

        rather than starting with a "Login with...".

        To work, `.login_url()` must give a URL other than the default,
        such as an oauth handler or another automatic login handler,
        registered with `.get_handlers()`.

        """,
    )

    auto_login_oauth2_authorize = Bool(
        False,
        config=True,
        help="""
        Automatically begin login process for OAuth2 authorization requests

        When another application is using Grader Service as OAuth2 provider, it
        sends users to `services/grader/api/oauth2/authorize`. If the user isn't logged
        in already, and auto_login is not set, the user will be dumped on the
        hub's home page, without any context on what to do next.

        Setting this to true will automatically redirect users to login if
        they aren't logged in *only* on the `services/grader/api/oauth2/authorize`
        endpoint.

        """,
    )

    def login_url(self, base_url):
        """Override this when registering a custom login handler

        Generally used by authenticators that do not use simple form-based authentication.

        The subclass overriding this is responsible for making sure there is a handler
        available to handle the URL returned from this method, using the `get_handlers`
        method.

        Args:
            base_url (str): the base URL of the Grader Service (e.g. /services/grader)

        Returns:
            str: The login URL, e.g. '/services/grader/login'
        """
        return url_path_join(base_url, "login")

    def logout_url(self, base_url):
        """Override when registering a custom logout handler

        The subclass overriding this is responsible for making sure there is a handler
        available to handle the URL returned from this method, using the `get_handlers`
        method.

        Args:
            base_url (str): the base URL of the Hub (e.g. /hub/)

        Returns:
            str: The logout URL, e.g. '/hub/logout'
        """
        return url_path_join(base_url, "logout")

    def get_handlers(self, base_url: str):
        """Return any custom handlers the authenticator needs to register

        Used in conjugation with `login_url` and `logout_url`.

        Args:
            base_url (str):
                the application object, in case it needs to be accessed for info.
        Returns:
            handlers (list):
                list of ``('/url', Handler)`` tuples passed to tornado.
                The Hub prefix is added to any URLs.
        """
        return [(self.login_url(base_url), LoginHandler)]



import json
import os

from tornado.escape import json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from traitlets import Unicode

from grader_service.auth.auth import Authenticator
from grader_service.auth.login import LoginHandler, LogoutHandler
from grader_service.orm.api_token import APIToken


class JupyterHubTokenAuthenticator(Authenticator):
    user_info_url = Unicode(
        config=True,
        help="""The URL to where this authenticator makes a request to acquire user
        details with an access token received via jupyterhub.""",
    )

    http_client = AsyncHTTPClient()

    async def authenticate(self, handler, data):
        headers = {"Authorization": f"Bearer {data['token']}"}
        request = HTTPRequest(url=self.user_info_url, headers=headers, method="GET")

        response = await self.http_client.fetch(request=request)
        response = json_decode(response.body)
        username = response["name"]
        display_name = response.get("display_name")
        groups = response["groups"]

        return {"name": username, "display_name": display_name, "groups": groups}

    def get_handlers(self, base_url):
        handlers = [
            (self.logout_url(base_url), LogoutHandler),
            (self.login_url(base_url), TokenLoginHandler),
        ]
        return handlers


class TokenLoginHandler(LoginHandler):
    async def post(self):
        data = json.loads(self.request.body)
        user = await self.login_user(data)

        if user:
            # create a API token for the user, that he can use to authenticate and return it
            token = APIToken.new(
                user=user,
                scopes=["identify"],  # Define the scopes for the token
                note="User login token",
                expires_in=os.environ.get("TOKEN_EXPIRES_IN", 1209600),  # Set expiration time
            )
            self.write({"api_token": token})
        else:
            html = await self._render(login_error="Invalid username or password")
            self.set_status(404)
            await self.finish(html)



from datetime import datetime, timezone

from sqlalchemy import Column, ForeignKey, Integer, Unicode, or_
from sqlalchemy.orm import joinedload, relationship

from grader_service.orm.base import Base
from grader_service.orm.json_util import JSONList


class OAuthCode(Base):
    __tablename__ = "oauth_code"
    id = Column(Integer, primary_key=True, autoincrement=True)
    client_id = Column(Unicode(255), ForeignKey("oauth_client.identifier", ondelete="CASCADE"))
    client = relationship("OAuthClient", back_populates="codes")
    code = Column(Unicode(36))
    expires_at = Column(Integer)
    redirect_uri = Column(Unicode(1023))
    session_id = Column(Unicode(255))
    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"))
    user = relationship("User", back_populates="oauth_codes")

    scopes = Column(JSONList, default=[])

    @staticmethod
    def now():
        return datetime.now(timezone.utc).timestamp()

    @classmethod
    def find(cls, db, code):
        return (
            db.query(cls)
            .filter(cls.code == code)
            .filter(or_(cls.expires_at is None, cls.expires_at >= cls.now()))
            .options(
                # load user with the code
                joinedload(cls.user, innerjoin=True)
            )
            .first()
        )

    def __repr__(self):
        return f"<{self.__class__.__name__}(id={self.id}, client_id={self.client_id!r})>"



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Integer

from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class LimitOutput(NbGraderPreprocessor):
    """Preprocessor for limiting cell output"""

    max_lines = Integer(1000, help="maximum number of lines of output (-1 means no limit)").tag(
        config=True
    )

    max_traceback = Integer(100, help="maximum number of traceback lines (-1 means no limit)").tag(
        config=True
    )

    def _limit_stream_output(self, cell: NotebookNode) -> NotebookNode:
        if self.max_lines == -1 or cell.cell_type != "code":
            return cell

        length = 0
        new_outputs = []
        for output in cell.outputs:
            if output.output_type == "stream":
                if length == self.max_lines:
                    continue

                text = output.text.split("\n")
                if (len(text) + length) > self.max_lines:
                    text = text[: (self.max_lines - length - 1)]
                    text.append("... Output truncated ...")

                length += len(text)
                output.text = "\n".join(text)

            new_outputs.append(output)

        cell.outputs = new_outputs
        return cell

    def _limit_traceback(self, cell: NotebookNode) -> NotebookNode:
        if self.max_traceback == -1 or cell.cell_type != "code":
            return cell

        for output in cell.outputs:
            if output.output_type == "error":
                if len(output.traceback) > self.max_traceback:
                    start = int(self.max_traceback / 2)
                    end = self.max_traceback - start - 1
                    tb = output.traceback[:start]
                    tb.append("... Traceback truncated ...")
                    tb.extend(output.traceback[-end:])
                    output.traceback = tb

        return cell

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        cell = self._limit_stream_output(cell)
        cell = self._limit_traceback(cell)
        return cell, resources



import re
from textwrap import dedent
from typing import Any, Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Dict, Unicode
from traitlets.config.loader import Config

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ClearSolutions(NbGraderPreprocessor):
    code_stub = Dict(
        default_value={
            "python": "# YOUR CODE HERE\nraise NotImplementedError()",
            "matlab": "% YOUR CODE HERE\nerror('No Answer Given!')",
            "octave": "% YOUR CODE HERE\nerror('No Answer Given!')",
            "sas": "/* YOUR CODE HERE */\n %notImplemented;",
            "java": "// YOUR CODE HERE",
            "C++11": "// YOUR CODE HERE",
            "C++14": "// YOUR CODE HERE",
            "C++17": "// YOUR CODE HERE",
            "r": "# YOUR CODE HERE\nfail() # No Answer - remove if you provide an answer",
            "R": "# YOUR CODE HERE\nfail() # No Answer - remove if you provide an answer",
        },
        help="The code snippet that will replace code solutions",
    ).tag(config=True)

    text_stub = Unicode(
        "YOUR ANSWER HERE", help="The text snippet that will replace written solutions"
    ).tag(config=True)

    begin_solution_delimeter = Unicode(
        "BEGIN SOLUTION", help="The delimiter marking the beginning of a solution"
    ).tag(config=True)

    end_solution_delimeter = Unicode(
        "END SOLUTION", help="The delimiter marking the end of a solution"
    ).tag(config=True)

    enforce_metadata = Bool(
        True,
        help=dedent(
            """
            Whether or not to complain if cells containing solutions regions are
            not marked as solution cells. WARNING: this will potentially cause
            things to break if you are using the full nbgrader pipeline. ONLY
            disable this option if you are only ever planning to use nbgrader
            assign.
            """
        ),
    ).tag(config=True)

    def _load_config(self, cfg: Config, **kwargs: Any) -> None:
        if "code_stub" in cfg.ClearSolutions:
            if not isinstance(cfg.ClearSolutions.code_stub, dict):
                self.log.warning(
                    "The ClearSolutions.code_stub option must now be given as a "
                    "dictionary with keys for the language of the notebook. I will "
                    "automatically convert ClearSolutions.code_stub to a dictionary "
                    "with a key for 'python', but note that this functionality may "
                    "be removed in future releases."
                )
                cfg.ClearSolutions.code_stub = dict(python=cfg.ClearSolutions.code_stub)

        if "comment_mark" in cfg.ClearSolutions:
            self.log.warning(
                "The ClearSolutions.comment_mark config option is deprecated. "
                "Please include the comment mark in ClearSolutions.begin_solution_delimeter "
                "and ClearSolutions.end_solution_delimeter instead."
            )
            del cfg.ClearSolutions.comment_mark

        super(ClearSolutions, self)._load_config(cfg, **kwargs)

    def _replace_solution_region(self, cell: NotebookNode, language: str) -> bool:
        """Find a region in the cell that is delimeted by
        `self.begin_solution_delimeter` and `self.end_solution_delimeter` (e.g.
        ### BEGIN SOLUTION and ### END SOLUTION). Replace that region either
        with the code stub or text stub, depending the cell type.

        This modifies the cell in place, and then returns True if a
        solution region was replaced, and False otherwise.

        """
        # pull out the cell input/source
        lines = cell.source.split("\n")
        if cell.cell_type == "code":
            stub_lines = self.code_stub[language].split("\n")
        else:
            stub_lines = self.text_stub.split("\n")

        new_lines = []
        in_solution = False
        replaced_solution = False

        for line in lines:
            # begin the solution area
            if self.begin_solution_delimeter in line:
                # check to make sure this isn't a nested BEGIN
                # SOLUTION region
                if in_solution:
                    raise RuntimeError("encountered nested begin solution statements")

                in_solution = True
                replaced_solution = True

                # replace it with the stub, indented as necessary
                indent = re.match(r"\s*", line).group(0)
                for stub_line in stub_lines:
                    new_lines.append(indent + stub_line)

            # end the solution area
            elif self.end_solution_delimeter in line:
                in_solution = False

            # add lines as long as it's not in the solution area
            elif not in_solution:
                new_lines.append(line)

        # we finished going through all the lines, but didn't find a
        # matching END SOLUTION statment
        if in_solution:
            raise RuntimeError("no end solution statement found")

        # replace the cell source
        cell.source = "\n".join(new_lines)

        return replaced_solution

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        language = nb.metadata.get("kernelspec", {}).get("language", "python")
        if language not in self.code_stub:
            raise ValueError(
                "language '{}' has not been specified in ClearSolutions.code_stub".format(language)
            )

        resources["language"] = language
        nb, resources = super(ClearSolutions, self).preprocess(nb, resources)
        if "celltoolbar" in nb.metadata:
            del nb.metadata["celltoolbar"]
        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # replace solution regions with the relevant stubs
        language = resources["language"]
        replaced_solution = self._replace_solution_region(cell, language)

        # determine whether the cell is a solution/grade cell
        is_solution = utils.is_solution(cell)

        # check that it is marked as a solution cell if we replaced a solution
        # region -- if it's not, then this is a problem, because the cell needs
        # to be given an id
        if not is_solution and replaced_solution:
            if self.enforce_metadata:
                raise RuntimeError(
                    "Solution region detected in a non-solution cell; please make sure "
                    "all solution regions are within solution cells."
                )

        # replace solution cells with the code/text stub -- but not if
        # we already replaced a solution region, because that means
        # there are parts of the cells that should be preserved
        if is_solution and not replaced_solution:
            if cell.cell_type == "code":
                cell.source = self.code_stub[language]
            else:
                cell.source = self.text_stub

        return cell, resources
